/*
*########################################################################
*########################################################################
*########################################################################
*                                                               
*   COPYRIGHT (c) 1998, 1999 by TransNexus, LLC                          
*                                                                    
*   This software contains proprietary and confidential information  
*   of TransNexus, LLC. Except as may be set forth in the license    
*   agreement under which this software is supplied, use, disclosure, 
*   or reproduction is prohibited without the prior, express, written
*   consent of TransNexus, LLC.                                      
*                                     
*
*
*
*
*
*
*#########################################################################
*#########################################################################
*#########################################################################
*/


/*
 * ospdest.h - OSP destination objects
 */

#ifndef _OSPDEST_H
#define _OSPDEST_H

#include "osp.h"
#include "ospxmlelem.h"
#include "ospmsg.h"
#include "ospcallid.h"
#include "osptoken.h"


typedef struct
{
    OSPTLISTLINK      ospmDestLink;
    unsigned char     ospmDestNumber[OSPC_E164NUMSIZE];
    unsigned char     ospmDestAddr[OSPC_SIGNALADDRSIZE];
    unsigned char     ospmDestDevAddr[OSPC_SIGNALADDRSIZE];
    OSPTTIME          ospmDestValidAfter;
    OSPTTIME          ospmDestValidUntil;
    OSPTLIST          ospmDestTokens;
    unsigned char     ospmDestAuthority[OSPC_URLSIZE];
    unsigned          ospmDestHasLimit;
    unsigned          ospmDestLimit;
    OSPTCALLID       *ospmDestCallId;
    unsigned          ospmDestTNFailReason;
    unsigned          ospmDestTNFailReasonInd;

}
OSPTDEST;


/**/
/*-----------------------------------------------------------------------*
 * macros that emulate functions
 *-----------------------------------------------------------------------*/
#ifndef OSPC_DEBUG
/*
 * Note: all macros are also implemented as functions in ospmsgdest.c.
 * For implementation details, see the comments in that file. To replace
 * a macro with a true function, simply comment out the macro definition
 * below.
 */

#define OSPPDestHasNumber(ospvDest) \
    ((ospvDest)->ospmDestNumber[0] != '\0')
#define OSPPDestSetNumber(ospvDest,ospvNum) \
            OSPM_MEMCPY((ospvDest)->ospmDestNumber, \
                (ospvNum), \
                min(OSPC_E164NUMSIZE,OSPM_STRLEN((const char *)ospvNum)+1))
#define OSPPDestGetNumber(ospvDest) ((ospvDest)->ospmDestNumber)

#define OSPPDestHasAddr(ospvDest) \
        ((ospvDest)->ospmDestAddr[0] != '\0')
#define OSPPDestSetAddr(ospvDest,ospvAddr) \
            OSPM_MEMCPY((ospvDest)->ospmDestAddr, (ospvAddr), \
                min(OSPC_SIGNALADDRSIZE,OSPM_STRLEN((const char *)ospvAddr)+1))
#define OSPPDestGetAddr(ospvDest) ((ospvDest)->ospmDestAddr)

#define OSPPDestDevHasAddr(ospvDest) \
        ((ospvDest)->ospmDestDevAddr[0] != '\0')
#define OSPPDestDevSetAddr(ospvDest,ospvAddr) \
            OSPM_MEMCPY((ospvDest)->ospmDestDevAddr, (ospvAddr), \
                min(OSPC_SIGNALADDRSIZE,OSPM_STRLEN((const char *)ospvAddr)+1))
#define OSPPDestDevGetAddr(ospvDest) ((ospvDest)->ospmDestDevAddr)


#define OSPPDestHasValidAfter(ospvDest) \
    ((ospvDest)->ospmDestValidAfter != OSPC_TIMEMIN)
#define OSPPDestSetValidAfter(ospvDest,ospvTime) \
    (ospvDest)->ospmDestValidAfter = (ospvTime)
#define OSPPDestGetValidAfter(ospvDest) (ospvDest)->ospmDestValidAfter

#define OSPPDestHasValidUntil(ospvDest) \
    ((ospvDest)->ospmDestValidUntil != OSPC_TIMEMAX)
#define OSPPDestSetValidUntil(ospvDest,ospvTime) \
    (ospvDest)->ospmDestValidUntil = (ospvTime)
#define OSPPDestGetValidUntil(ospvDest) (ospvDest)->ospmDestValidUntil

#define OSPPDestHasAuthority(ospvDest) \
    ((ospvDest)->ospmDestAuthority[0] != '\0')
#define OSPPDestSetAuthority(ospvDest,ospvAuth) \
    OSPM_STRNCPY((char *)(ospvDest)->ospmDestAuthority, (ospvAuth), \
    min(OSPM_STRLEN((const char *)ospvAuth)+1,OSPC_URLSIZE-1))

#define OSPPDestHasCallId(ospvDest) \
    ((ospvDest)->ospmDestCallId != OSPC_OSNULL)
#define OSPPDestGetCallId(ospvDest) ((ospvDest)->ospmDestCallId)

#define OSPPDestHasToken(ospvDest) (OSPPListFirst(&(ospvDest)->ospmDestTokens) != OSPC_OSNULL)
#define OSPPDestAddToken(ospvDest, ospvToken)   \
    OSPPListAppend(&(ospvDest)->ospmDestTokens, (ospvToken))
#define OSPPDestFirstToken(ospvDest) \
    OSPPListFirst(&(ospvDest)->ospmDestTokens)
#define OSPPDestNextToken(ospvDest, ospvToken) \
    OSPPListNext(&(ospvDest)->ospmDestTokens, (ospvToken))

#define OSPPDestHasLimit(ospvDest) (((OSPTDEST *)ospvDest)->ospmDestHasLimit)
#define OSPPDestGetLimit(ospvDest) (((OSPTDEST *)ospvDest)->ospmDestLimit)
#define OSPPDestSetLimit(ospvDest, ospvLimit) \
    { ((OSPTDEST *)ospvDest)->ospmDestLimit=ospvLimit; \
    ((OSPTDEST *)ospvDest)->ospmDestHasLimit=OSPC_TRUE; }


/* Failure Reason */
#define OSPPDestHasTNFailReason(ospvDest) \
    ((ospvDest)->ospmDestTNFailReasonInd)
#define OSPPDestSetTNFailReason(ospvDest,ospvTNFailReason) { \
    (ospvDest)->ospmDestTNFailReason = (ospvTNFailReason); \
    (ospvDest)->ospmDestTNFailReasonInd = 1; \
    }
#define OSPPDestGetTNFailReason(ospvDest) \
    (ospvDest)->ospmDestTNFailReason

#endif /* OSPC_DEBUG */

#ifdef __cplusplus
extern "C" 
{
#endif

    /**/
    /*-----------------------------------------------------------------------*
     * function prototypes
     *-----------------------------------------------------------------------*/
    OSPTDEST      *OSPPDestNew(void);
    void           OSPPDestDelete(OSPTDEST **);
    unsigned       OSPPDestFromElement(OSPTXMLELEM *, OSPTDEST **);
    void           OSPPDestSetCallId(OSPTDEST *, const unsigned char *, unsigned);

#ifdef OSPC_DEBUG
    unsigned       OSPPDestHasNumber(OSPTDEST *ospvDest);
    void           OSPPDestSetNumber(OSPTDEST *, const unsigned char *);
    unsigned char *OSPPDestGetNumber(OSPTDEST *);

    unsigned       OSPPDestHasAddr(OSPTDEST *);
    void           OSPPDestSetAddr(OSPTDEST *, const unsigned char *);
    unsigned char *OSPPDestGetAddr(OSPTDEST *);

    unsigned       OSPPDestDevHasAddr(OSPTDEST *);
    void           OSPPDestDevSetAddr(OSPTDEST *, const unsigned char *);
    unsigned char *OSPPDestDevGetAddr(OSPTDEST *);

    unsigned       OSPPDestHasValidAfter(OSPTDEST *);
    void           OSPPDestSetValidAfter(OSPTDEST *, OSPTTIME);
    OSPTTIME       OSPPDestGetValidAfter(OSPTDEST *);

    unsigned       OSPPDestHasValidUntil(OSPTDEST *);
    void           OSPPDestSetValidUntil(OSPTDEST *, OSPTTIME);
    OSPTTIME       OSPPDestGetValidUntil(OSPTDEST *);

    unsigned       OSPPDestHasAuthority(OSPTDEST *);
    void           OSPPDestSetAuthority(OSPTDEST *, const unsigned char *);

    unsigned       OSPPDestHasCallId(OSPTDEST *);
    OSPTCALLID    *OSPPDestGetCallId(OSPTDEST *);

    unsigned       OSPPDestHasToken(OSPTDEST *);
    void           OSPPDestAddToken(OSPTDEST *, OSPTTOKEN *);
    OSPTTOKEN     *OSPPDestFirstToken(OSPTDEST *);
    OSPTTOKEN     *OSPPDestNextToken(OSPTDEST *, OSPTTOKEN *);

    unsigned       OSPPDestHasLimit(OSPTDEST *);
    unsigned       OSPPDestGetLimit(OSPTDEST *);
    void           OSPPDestSetLimit(OSPTDEST *, unsigned);

    unsigned       OSPPDestHasTNFailReason(OSPTDEST *);
    void           OSPPDestSetTNFailReason(OSPTDEST *, unsigned);
    unsigned       OSPPDestGetTNFailReason(OSPTDEST *);

#endif /* OSPC_DEBUG */

#ifdef __cplusplus
}
#endif

#endif /* _OSPDEST_H */

