/*
*########################################################################
*########################################################################
*########################################################################
*                                                               
*   COPYRIGHT (c) 1998, 1999 by TransNexus, LLC                          
*                                                                    
*   This software contains proprietary and confidential information  
*   of TransNexus, LLC. Except as may be set forth in the license    
*   agreement under which this software is supplied, use, disclosure, 
*   or reproduction is prohibited without the prior, express, written
*   consent of TransNexus, LLC.                                      
*                                     
*
*
*
*
*
*
*#########################################################################
*#########################################################################
*#########################################################################
*/


/*
 * ospusageind.h - OSP Usage Indication Objects
 */

#ifndef _OSPUSAGEIND_H
#define _OSPUSAGEIND_H

#include "osp.h"
#include "ospxmlelem.h"
#include "ospmsg.h"
#include "ospcallid.h"
#include "ospaltinfo.h"
#include "ospstatistics.h"

typedef struct
{
    OSPTLISTLINK      ospmUsageIndLink;
    OSPTTIME          ospmUsageIndTimestamp;
    unsigned char     *ospmUsageIndComponentId;
    unsigned          ospmUsageIndRole;
    unsigned          ospmUsageIndHasRole;
    OSPTTRXID         ospmUsageIndTransactionId;
    OSPTCALLID        *ospmUsageIndCallId;
    int               ospmUsageIndDuration;
    unsigned char     ospmUsageIndSourceNumber[OSPC_E164NUMSIZE];
    unsigned char     ospmUsageIndDestNumber[OSPC_E164NUMSIZE];
    OSPTLIST          ospmUsageIndSourceAlternate;
    OSPTLIST          ospmUsageIndDestinationAlternate;
    unsigned long     ospmUsageIndTNCustId;
    unsigned long     ospmUsageIndTNDeviceId;
    unsigned          ospmUsageIndTNFailReason;
    unsigned          ospmUsageIndTNFailReasonInd;
    OSPTSTATISTICS    *ospmUsageIndTNStats;
}
OSPTUSAGEIND;

/**/
/*-----------------------------------------------------------------------*
 * macros that emulate functions
 *-----------------------------------------------------------------------*/
#ifndef OSPC_DEBUG
/*
 * Note: all macros are also implemented as functions in ospusageind.c.
 * For implementation details, see the comments in that file. To replace
 * a macro with a true function, simply comment out the macro definition
 * below.
 */

/* Timestamp */
#define OSPPUsageIndHasTimestamp(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTimestamp)
#define OSPPUsageIndSetTimestamp(ospvUsageInd,ospvTime) \
    (ospvUsageInd)->ospmUsageIndTimestamp = (ospvTime)
#define OSPPUsageIndGetTimestamp(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndTimestamp

/* ComponentId */
#define OSPPUsageIndHasComponentId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndComponentId != OSPC_OSNULL)
#define OSPPUsageIndGetComponentId(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndComponentId

/* Role */
#define OSPPUsageIndHasRole(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndRole)
#define OSPPUsageIndGetRole(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndRole
#define OSPPUsageIndSetRole(ospvUsageInd,ospvRole) \
    (ospvUsageInd)->ospmUsageIndRole = (ospvRole); \
    (ospvUsageInd)->ospmUsageIndHasRole = OSPC_TRUE;

/* TransactionId */
#define OSPPUsageIndHasTransactionId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTransactionId != 0)
#define OSPPUsageIndSetTransactionId(ospvUsageInd,ospvTransactionId) \
    (ospvUsageInd)->ospmUsageIndTransactionId = ospvTransactionId
#define OSPPUsageIndGetTransactionId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTransactionId)

/* CallId */
#define OSPPUsageIndHasCallId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndCallId != OSPC_OSNULL)
#define OSPPUsageIndGetCallId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndCallId)

/* SourceInfo */
#define OSPPUsageIndSetSourceNumber(ospvUsageInd,ospvSourceNumber) \
    OSPM_STRNCPY((char *)(ospvUsageInd)->ospmUsageIndSourceNumber, \
        (const char *)(ospvSourceNumber), \
        min(OSPC_E164NUMSIZE-1, OSPM_STRLEN((const char *) ospvSourceNumber)+1))
#define OSPPUsageIndGetSourceNumber(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndSourceNumber)

/* Source Alternate */
#define OSPPUsageIndHasSourceAlt(ospvUsageInd) \
    (OSPPUsageIndFirstSourceAlt(ospvUsageInd) != OSPC_OSNULL)

#define OSPPUsageIndFirstSourceAlt(ospvUsageInd) \
    (const char *)OSPPListFirst(&((ospvUsageInd)->ospmUsageIndSourceAlternate))

#define OSPPUsageIndNextSourceAlt(ospvUsageInd, ospvAltInfo) \
    (OSPTALTINFO *)OSPPListNext((OSPTLIST *)&((ospvUsageInd)->ospmUsageIndSourceAlternate), \
                        (void *)ospvAltInfo)

/* DestinationInfo */
#define OSPPUsageIndSetDestNumber(ospvUsageInd,ospvDestNumber) \
    OSPM_STRNCPY((char *)(ospvUsageInd)->ospmUsageIndDestNumber, \
        (const char *)(ospvDestNumber),  \
        min(OSPC_E164NUMSIZE-1, OSPM_STRLEN((const char *) ospvDestNumber)+1))
#define OSPPUsageIndGetDestNumber(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndDestNumber)

/* Destination Alternate */

#define OSPPUsageIndHasDestinationAlt(ospvUsageInd) \
    (OSPPUsageIndFirstDestinationAlt(ospvUsageInd) != OSPC_OSNULL)

#define OSPPUsageIndFirstDestinationAlt(ospvUsageInd) \
    (const char *)OSPPListFirst(&((ospvUsageInd)->ospmUsageIndDestinationAlternate))

#define OSPPUsageIndNextDestinationAlt(ospvUsageInd, ospvAltInfo) \
    (OSPTALTINFO *)OSPPListNext((OSPTLIST *)&((ospvUsageInd)->ospmUsageIndDestinationAlternate), \
                        (void *)ospvAltInfo)

#define OSPPUsageIndGetDestinationAltSize(ospvAltInfo) \
    OSPPAltInfoGetSize(ospvAltInfo)

/* UsageDetail */
#define OSPPUsageIndHasDuration(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndDuration >= 0)
#define OSPPUsageIndSetDuration(ospvUsageInd,ospvDuration) \
    (ospvUsageInd)->ospmUsageIndDuration = ospvDuration;
#define OSPPUsageIndGetDuration(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndDuration)

/* Transnexus Extensions */
/* Customer ID */
#define OSPPUsageIndHasTNCustId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTNCustId != 0L)
#define OSPPUsageIndSetTNCustId(ospvUsageInd,ospvTNCustId) \
    (ospvUsageInd)->ospmUsageIndTNCustId = (ospvTNCustId)
#define OSPPUsageIndGetTNCustId(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndTNCustId

/* Device ID */
#define OSPPUsageIndHasTNDeviceId(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTNDeviceId != 0L)
#define OSPPUsageIndSetTNDeviceId(ospvUsageInd,ospvTNDeviceId) \
    (ospvUsageInd)->ospmUsageIndTNDeviceId = (ospvTNDeviceId)
#define OSPPUsageIndGetTNDeviceId(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndTNDeviceId

/* Failure Reason */
#define OSPPUsageIndHasTNFailReason(ospvUsageInd) \
    ((ospvUsageInd)->ospmUsageIndTNFailReasonInd)
#define OSPPUsageIndSetTNFailReason(ospvUsageInd,ospvTNFailReason) { \
    (ospvUsageInd)->ospmUsageIndTNFailReason = (ospvTNFailReason); \
    (ospvUsageInd)->ospmUsageIndTNFailReasonInd = 1; \
    }
#define OSPPUsageIndGetTNFailReason(ospvUsageInd) \
    (ospvUsageInd)->ospmUsageIndTNFailReason

#endif

/**/
/*-----------------------------------------------------------------------*
 * function prototypes
 *-----------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" 
{
#endif

    OSPTUSAGEIND    *OSPPUsageIndNew(void);
    void            OSPPUsageIndDelete( OSPTUSAGEIND **);
    int             OSPPUsageIndToElement(OSPTLIST *, OSPTXMLELEM **);
    void            OSPPUsageIndSetCallId(OSPTUSAGEIND *, OSPTCALLID *);

    void      OSPPUsageIndMoveSourceAlt(OSPTUSAGEIND *, OSPTLIST *);
    void      OSPPUsageIndMoveDestinationAlt(OSPTUSAGEIND *, OSPTLIST *);
    void      OSPPUsageIndCopySourceAlt(OSPTUSAGEIND *, OSPTLIST *);
    void      OSPPUsageIndSetComponentId(OSPTUSAGEIND *, unsigned char *);

#ifdef OSPC_DEBUG

    unsigned       OSPPUsageIndHasTimestamp(OSPTUSAGEIND *);
    void           OSPPUsageIndSetTimestamp(OSPTUSAGEIND *, OSPTTIME);
    OSPTTIME       OSPPUsageIndGetTimestamp(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasComponentId(OSPTUSAGEIND *);
    unsigned char  *OSPPUsageIndGetComponentId(OSPTUSAGEIND *);

    unsigned            OSPPUsageIndHasRole(OSPTUSAGEIND *);
    OSPE_MSG_ROLETYPES  OSPPUsageIndGetRole(OSPTUSAGEIND *);
    void                OSPPUsageIndSetRole(OSPTUSAGEIND *, unsigned);

    unsigned       OSPPUsageIndHasTransactionId(OSPTUSAGEIND *);
    void           OSPPUsageIndSetTransactionId(OSPTUSAGEIND *, OSPTTRXID ospvTransactionId);
    OSPTTRXID      OSPPUsageIndGetTransactionId(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasCallId(OSPTUSAGEIND *);
    OSPTCALLID    *OSPPUsageIndGetCallId(OSPTUSAGEIND *);

    void           OSPPUsageIndSetSourceNumber(OSPTUSAGEIND *,unsigned char *);
    unsigned char *OSPPUsageIndGetSourceNumber(OSPTUSAGEIND *);

    void           OSPPUsageIndSetDestNumber(OSPTUSAGEIND *,unsigned char *);
    unsigned char *OSPPUsageIndGetDestNumber(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasSourceAlt(OSPTUSAGEIND *);
    OSPTALTINFO   *OSPPUsageIndFirstSourceAlt(OSPTUSAGEIND *);
    OSPTALTINFO   *OSPPUsageIndNextSourceAlt(OSPTUSAGEIND *, OSPTALTINFO  *);

    unsigned       OSPPUsageIndHasDestinationAlt(OSPTUSAGEIND *);
    unsigned       OSPPUsageIndGetDestinationAltSize(OSPTALTINFO *);
    OSPTALTINFO   *OSPPUsageIndFirstDestinationAlt(OSPTUSAGEIND *);
    OSPTALTINFO   *OSPPUsageIndNextDestinationAlt(OSPTUSAGEIND *, OSPTALTINFO  *);

    unsigned       OSPPUsageIndHasDuration(OSPTUSAGEIND *);
    void           OSPPUsageIndSetDuration(OSPTUSAGEIND *, int ospvDuration);
    int            OSPPUsageIndGetDuration(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasTNCustId(OSPTUSAGEIND *ospvUsageInd);
    void           OSPPUsageIndSetTNCustId(OSPTUSAGEIND *, unsigned long);
    unsigned long  OSPPUsageIndGetTNCustId(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasTNDeviceId(OSPTUSAGEIND *ospvUsageInd);
    void           OSPPUsageIndSetTNDeviceId(OSPTUSAGEIND *, unsigned long);
    unsigned long  OSPPUsageIndGetTNDeviceId(OSPTUSAGEIND *);

    unsigned       OSPPUsageIndHasTNFailReason(OSPTUSAGEIND *ospvUsageInd);
    void           OSPPUsageIndSetTNFailReason(OSPTUSAGEIND *, unsigned);
    unsigned       OSPPUsageIndGetTNFailReason(OSPTUSAGEIND *);


#endif /* OSPC_DEBUG */

    OSPTBOOL       OSPPUsageIndHasTNStatistics(OSPTUSAGEIND *);
    void           OSPPUsageIndSetTNStatistics(OSPTUSAGEIND *, OSPTSTATISTICS *);

#ifdef __cplusplus
}
#endif

#endif /* _OSPUSAGEIND_H */
