#ifndef __LIBROUTE_H__
#define __LIBROUTE_H__ 1

/*
 * libroute.h Interface to Linux routing engine.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 */

#define RTO_NOUPDATE	1	/* Only test cache */
#define RTO_NOWAIT	2	/* Do not wait for result */

#define R_NONE		0
#define R_FAILED	1
#define R_INCOMPLETE	2
#define R_PROBE		4
#define R_STALE		8
#define R_FRESH		0x10
#define R_VALID		(R_FRESH|R_STALE|R_PROBE)

#ifdef USE_IPV6
#define KEYLEN		4
#else
#define KEYLEN		1
#endif

struct dstentry
{
	struct dstentry	*next;
	struct dstentry	*queue_next;

	unsigned char	family;
	unsigned char	state;
	unsigned char	block;
	unsigned char	refresh:1;

	time_t		updated;
	time_t		accessed;
	int		error;
	int		iif;
	__u32		seq;

	__u32		dst[KEYLEN];
	__u32		src[KEYLEN];

	struct rtattr	*path;
};

extern int netlink_read_wakeup(void);
extern struct dstentry * dst_route(int family, void *dst, void *src, int iif, unsigned flags);
extern int dst_init(void (*refresh)(struct dstentry*));
extern int dst_is_connected(struct dstentry *d);
extern int dst_oif(struct dstentry *d);

#endif /* __LIBROUTE_H__ */
