/*
 *  @(#) $Id: rsrr_unicast.c,v 1.1.1.1 2000/05/08 22:51:24 wenqing Exp $
 */

/************************ rsvp_unicast.c  ****************************
 *                                                                   *
 *       System-dependent Unicast route lookup routine               *
 *                                                                   *
 *********************************************************************/

#define	USYNCH 1

#include "rsvp_daemon.h"
#include <sys/param.h>

#ifdef sun
#ifdef SOLARIS
#include <sys/stream.h>
#include <inet/common.h>
#undef IPOPT_EOL
#undef IPOPT_NOP
#undef IPOPT_LSRR
#undef IPOPT_RR
#undef IPOPT_SSRR
#include <inet/ip.h>
#else /* not SOLARIS */
#include <sys/mbuf.h>
#endif /* SOLARIS */
#endif /* sun */

u_long bmp_unicast;
int unicast_result_ready;
int unicast_query_waiting;
int asynchronous = 1;

static int init_failed = 0;
int locate_LIH(unsigned int, net_addr *);

int
locate_LIH(unsigned int ifidx, net_addr *addr)
{
	int i;
	for (i = 0; i < if_num; i++) {
		if (IsNumAPI(i))
			continue;
		if (ifidx == if_vec[i].if_index) {
			if (net_addr_equal(addr, &(NET_GET_IF_PHY_ADDR(&(if_vec[i].if_addr)))))
				return i;
		}
	}
	return -1;
}

#ifdef SOLARIS
#include <sys/utsname.h>

#define SOLARIS_2_6	6

/*
 * Routing socket indicies
 */
#define RTAX_DST	0
#define RTAX_GATEWAY	1
#define RTAX_NETMASK	2
#define RTAX_GENMASK	3
#define RTAX_IFP	4
#define RTAX_IFA	5
#define RTAX_AUTHOR	6
#define RTAX_BRD	7
#define RTAX_MAX	8

static int
getSolarisRelease(void)
{
	struct utsname	n;
	int		major, minor;
	char		*period, *release;

	if (uname(&n) < 0) {
		perror("uname");
		return (-1);
	}

	release = strdup(n.release);

	if ((period = strchr(release, '.'))) {
		*period = '\0';
		minor = atoi(period + 1);
	}

	major = atoi(release);
	free (release);

	if (major != 5)
		return (-1);
	else
		return (minor);
}
#endif	/* SOLARIS */

#if defined(PF_ROUTE)  && !defined(sgi_53) && !defined(linux)

int seq;		/* routing socket sequence num */
int rskt;		/* routing socket */

/*
 * This code will assumes a routing socket and will work on many newer
 * BSD-like machines
 *
 */

/*
 * Copyright 1995 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that both the above copyright notice and this
 * permission notice appear in all copies, that both the above
 * copyright notice and this permission notice appear in all
 * supporting documentation, and that the name of M.I.T. not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  M.I.T. makes
 * no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THIS SOFTWARE IS PROVIDED BY M.I.T. ``AS IS''.  M.I.T. DISCLAIMS
 * ALL EXPRESS OR IMPLIED WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
 * SHALL M.I.T. BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <errno.h>
#include <net/if.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <netinet/in.h>

/*
 * Round up 'a' to next multiple of 'size'
 */ 
#define ROUNDUP(a, size) (((a) & ((size)-1)) ? (1 + ((a) | ((size)-1))) : (a))
  
/*
 * Step to next socket address structure;
 * if sa_len is 0, assume it is sizeof(u_long).
 */
#define NEXT_SA(ap) ap = (struct sockaddr *) \
    ((caddr_t) ap + (ap->sa_len ? ROUNDUP(ap->sa_len, sizeof (u_long)) : \
                                    sizeof(u_long)))

#ifdef SOLARIS
static int useRoutingSocket = 0;
#endif

/*
 * For now, we open a new socket every time we want to get a route, to
 * save some effort in parsing.  Eventually we should keep it open and
 * listen for changes.  For now, just open and close it to make sure that
 * we can.
 */
int
unicast_init(void)
{
#ifdef SOLARIS
	static int unicast_init_specific();
	int	release;
	if ((release = getSolarisRelease()) < 0) {
		init_failed = 1;
		NoUnicast = 1;
		return (0);
	}
	if (release >= SOLARIS_2_6) {
		rskt = socket(PF_ROUTE, SOCK_RAW, AF_INET);
		if (rskt < 0) {
			log(LOG_INFO, 0,
				"Unicast routing information unavailable\n");
			init_failed = 1;
			NoUnicast = 1;
			return (0);
		}
		useRoutingSocket++;
	} else if (!unicast_init_specific()) {
		init_failed = 1;
		NoUnicast = 1;
		return (0);
	}
#else	/* SOLARIS */
	rskt = socket(PF_ROUTE, SOCK_RAW, PF_UNSPEC);
	if (rskt < 0) {
		init_failed = 1;
		NoUnicast = 1;
		log(LOG_INFO, 0, "Unicast routing information unavailable\n");
		return 0;
	}
#endif	/* SOLARIS */
	NoUnicast = 0;
	close(rskt);
	return 1;
}

#define RTM_BUFLEN 2048		/* XXX should be in header file */
#define WORD_BNDARY(x)  (((x)+sizeof(long)-1) & ~(sizeof(long)-1))

#ifdef SBM
u_long
unicast_route_gw_routing_socket(u_long addr)
{
	int			s, one = 1, i, len;
	struct sockaddr		*sa;
#ifdef SOLARIS
	struct sockaddr		*rti_info[RTA_NUMBITS];
#else
	struct sockaddr		*rti_info[RTAX_MAX];
#endif
	struct sockaddr_in	*isa;
	struct rt_msghdr	*mhp;
	caddr_t			cp;
	char			buf[RTM_BUFLEN];

	if ((s = socket(PF_ROUTE, SOCK_RAW, AF_INET)) < 0) {
		perror("socket");
		return (-1);
	}

	setsockopt(s, SOL_SOCKET, SO_USELOOPBACK, (char *)&one, sizeof (one));

	/* Set up GET message */
	memset(buf, 0, sizeof buf);

	mhp = (struct rt_msghdr *)buf;
	cp = buf + sizeof(struct rt_msghdr);

	isa = (struct sockaddr_in *)cp;
	isa->sin_family = AF_INET;
	isa->sin_addr.s_addr = addr;

	mhp->rtm_version = RTM_VERSION;
	mhp->rtm_type = RTM_GET;
	mhp->rtm_addrs = RTA_GATEWAY; 
	mhp->rtm_seq = ++seq;
	mhp->rtm_msglen = sizeof(struct rt_msghdr) + sizeof(struct sockaddr_in);

	/* Send the message */
	if (write(s, buf, (size_t)mhp->rtm_msglen) < 0) {
		log(LOG_ERR, errno, "writing message on routing socket\n");
		close(s);
		return -1;
	}

	/* Read the reply */
	do {
		len = read(s, buf, sizeof buf);
	} while (len > 0 && mhp->rtm_seq != seq && mhp->rtm_pid != getpid());

	close(s);

	if (len < 0 || mhp->rtm_errno) {
		if (IsDebug(DEBUG_ROUTE))
			log(LOG_DEBUG, errno, "unicast route lookup failed");
		return -1;
	}

	if (len == 0) {
		log(LOG_ERR, 0, "unicast route lookup is confused");
		return -1;
	}
	sa = (struct sockaddr *) (mhp + 1);

	get_route_info(sa, rti_info, mhp);	

	isa = NULL;
	for (i = 0; i < RTAX_MAX; i++) {
		switch(mhp->rtm_addrs & (1 << i)) {
		case RTA_GATEWAY:
			isa = (struct sockaddr_in *)rti_info[i];
			break;
		}
	}

	if (!isa)
		return (-1);
	else
		return (isa->sin_addr.s_addr);
}

u_long
unicast_route_gw(u_long addr)
{
#ifdef SOLARIS
	if (useRoutingSocket)
		return unicast_route_gw_routing_socket(addr);
	else
		return unicast_route_gw_specific(addr);
#else
	unicast_route_gw_routing_socket(addr);
#endif
}

static void
get_route_info(struct sockaddr *sa, struct sockaddr **rti_info,
	struct rt_msghdr *mhp)
{
	int		i;

#ifdef SOLARIS
	caddr_t		cp;
	cp  = (caddr_t)sa;
	for (i = 0; i < RTA_NUMBITS; i++) {
		switch (mhp->rtm_addrs & (1 << i)) {
		case RTA_DST:
		case RTA_GATEWAY:
		case RTA_NETMASK:
		case RTA_GENMASK:
		case RTA_AUTHOR:
		case RTA_IFA:
			rti_info[i] = (struct sockaddr *)cp;
			cp += sizeof (struct sockaddr_in);
			break;

		case RTA_IFP:
			rti_info[i] = (struct sockaddr *)cp;
			cp += sizeof (struct sockaddr_dl);
			break;
		}
	}
#else
    	for (i = 0; i < RTAX_MAX; i++) {
        	if (mhp->rtm_addrs & (1 << i)) {
			rti_info[i] = sa;
			NEXT_SA(sa);
        	} else
			rti_info[i] = NULL;
    	}   
#endif
}

#endif

#ifdef SOLARIS
int
unicast_route(net_addr *addr)
{
	static int unicast_route_routing_socket(net_addr *);
	static int unicast_route_specific(u_long);

	switch (NET_GET_TYPE(addr)) {
	case NET_ADDR_IPv4:
		if (useRoutingSocket)
			return (unicast_route_routing_socket(addr));
		else
			return(unicast_route_specific(
				NET_GET_ADDR_IPv4(addr).s_addr));

	case NET_ADDR_UDP_IPv4:
		if (useRoutingSocket)
			return (unicast_route_routing_socket(addr));
		else
			return(unicast_route_specific(
				NET_GET_ADDR_UDP_IPv4(addr).sin_addr.s_addr));
	default:
		return(0);
	}

}

#endif	/* SOLARIS */

/*
 * Find the kernel's idea of the route to destination DEST (in host order)
 */
int
#ifdef SOLARIS
unicast_route_routing_socket(net_addr *addr)
#else
unicast_route(net_addr *addr)
#endif
{
	struct sockaddr *sp;
	struct sockaddr_in *s4;
	struct rt_msghdr *mhp;
	net_addr *addr2;
	char buf[RTM_BUFLEN];
	int one, sz;
	int sock;
	int pid;
#ifdef  USYNCH
	int s;
	int len;
	struct sockaddr *sa;
#ifdef SOLARIS
	caddr_t		cp;
	struct sockaddr *rti_info[RTA_NUMBITS];
#else
	struct sockaddr *rti_info[RTAX_MAX];
#endif
    net_addr nadr;
    struct sockaddr_dl *sdlp;
    char ifname[IFNAMSIZ];
    int i;
    int ifidx = 0;
#endif  /* USYNCH */
#ifdef	USE_IPV6
	struct sockaddr_in6 *s6;
#endif	/* USE_IPV6 */


#ifdef	USYNCH
	/* Open socket */
	s = socket(PF_ROUTE, SOCK_RAW, PF_UNSPEC);
	if (s < 0) {
		log(LOG_ERR, errno,
		    "Could not open routing socket");
		return -1;
	}
	else 
		sock = s;
#else	/* USYNCH */
	sock = rskt;
#endif	/* !USYNCH */

	one = 1;
	setsockopt(sock, SOL_SOCKET, SO_USELOOPBACK, (char *) &one, sizeof one);

	/* Set up GET message */
	memset(buf, 0, sizeof buf);

	mhp = (struct rt_msghdr *)buf;
	sp = (struct sockaddr *)
               &buf[WORD_BNDARY(sizeof (struct rt_msghdr))];

	addr2 = net_addr_ip(addr);
	switch (NET_GET_TYPE(addr2)) {
		case NET_ADDR_IPv4:
			sz = sizeof(struct sockaddr_in);
			s4 = (struct sockaddr_in *) sp;
			s4->sin_family = AF_INET;
			s4->sin_addr.s_addr = NET_GET_ADDR_IPv4(addr2).s_addr;
			s4->sin_port = 0;
#ifdef SOCKADDR_LEN
			s4->sin_len = sz;
#endif
			break;
#ifdef	USE_IPV6
		case NET_ADDR_IPv6:
			sz = sizeof(struct sockaddr_in6);
			s6 = (struct sockaddr_in6 *) sp;
			s6->sin6_family = AF_INET6;
			s6->sin6_addr = NET_GET_ADDR_IPv6(addr2);
			s6->sin6_port = 0;
#ifdef SOCKADDR_LEN
			s6->sin6_len = sz;
#endif
			break;
#endif	/* USE_IPV6 */
		default:
			return(-1);
	}

	mhp->rtm_version = RTM_VERSION;
	mhp->rtm_type = RTM_GET;
	/* mhp->rtm_addrs = RTA_DST|RTA_IFP; */
	mhp->rtm_addrs = RTA_DST|RTA_IFP|RTA_IFA;
	mhp->rtm_seq = ++seq;
	mhp->rtm_pid = pid = getpid();
	mhp->rtm_msglen = ((char *)(sp)) + sz - buf;

	/* Send the message */
	if (write(sock, buf, (size_t)mhp->rtm_msglen) < 0) {
		log(LOG_ERR, errno, "writing message on routing socket\n");
#ifdef	USYNCH
		close(sock);
#endif	/* USYNCH */
		return -1;
	}

#if !defined(USYNCH)
	asynchronous = 1;
	return 0;	
#else	/* USYNCH */

	/* Read the reply */
	do {
		len = read(sock, buf, sizeof buf);
	} while (len > 0 && mhp->rtm_seq != seq && mhp->rtm_pid != pid);

	close(sock);

	if (len < 0 || mhp->rtm_errno) {
		if (IsDebug(DEBUG_ROUTE))
			log(LOG_DEBUG, errno, "unicast route lookup failed");
		return -1;
	}

	if (len == 0) {
		log(LOG_ERR, 0, "unicast route lookup is confused");
		return -1;
	}

	sa = (struct sockaddr *) (mhp + 1);
    
#ifdef SOLARIS
	cp  = (caddr_t)sa;
	for (i = 0; i < RTA_NUMBITS; i++) {
		switch (mhp->rtm_addrs & (1 << i)) {
		case RTA_DST:
		case RTA_GATEWAY:
		case RTA_NETMASK:
		case RTA_GENMASK:
		case RTA_AUTHOR:
		case RTA_IFA:
			rti_info[i] = (struct sockaddr *)cp;
			cp += sizeof (struct sockaddr_in);
			break;

		case RTA_IFP:
			rti_info[i] = (struct sockaddr *)cp;
			cp += sizeof (struct sockaddr_dl);
			break;
		}
	}
#else
    for (i = 0; i < RTAX_MAX; i++) {
        if (mhp->rtm_addrs & (1 << i)) {
            rti_info[i] = sa;
            NEXT_SA(sa);
        } else
            rti_info[i] = NULL;
    }   
#endif
 
	if ((sa = rti_info[RTAX_IFP]) != NULL) {
		sdlp = (struct sockaddr_dl *)sa;
		if (sdlp->sdl_family != AF_LINK || sdlp->sdl_nlen == 0) {
			if (IsDebug(DEBUG_ROUTE))
				log(LOG_DEBUG, 0, "unicast route through weird ifp");
			return -1;
		}
		strncpy(ifname, sdlp->sdl_data, sdlp->sdl_nlen);
		if (sdlp->sdl_nlen < IFNAMSIZ)
			ifname[sdlp->sdl_nlen] = '\0';
		ifidx = if_nametoindex(ifname);
	}

    if ((sa = rti_info[RTAX_IFA]) != NULL) {
	switch (sa->sa_family) {
		case AF_INET:
			s4 = (struct sockaddr_in *)sa;
			NET_SET_ADDR_IPv4(&nadr, s4->sin_addr);
			break;
#ifdef	USE_IPV6
		case AF_INET6:
			s6 = (struct sockaddr_in6 *)sa;
			NET_SET_ADDR_IPv6(&nadr, s6->sin6_addr);
			break;
#endif	/* USE_IPV6 */
		default:
			return (-1);
	}
    }
	else 
		return -1;

	asynchronous = 0;
	return (locate_LIH(ifidx, &nadr));
#endif	/* USYNCH */
}

# else

/* 
 * ifndef PF_ROUTE - no routing socket. OS-specific code required,
 */

#ifdef SOLARIS

static int unicast_init_specific();
static int unicast_route_specific(u_long);

int
unicast_init(void)
{
	return (unicast_init_specific());
}

#ifdef SBM

static int unicast_route_gw_specific(u_long);

u_long
unicast_route_gw(u_long addr)
{
	return (unicast_route_gw_specific(addr));
}
#endif

int
unicast_route(net_addr *addr)
{
	switch (NET_GET_TYPE(addr)) {
		case NET_ADDR_IPv4:
			return(unicast_route_specific(
				NET_GET_ADDR_IPv4(addr).s_addr));

		case NET_ADDR_UDP_IPv4:
			return(unicast_route_specific(
				NET_GET_ADDR_UDP_IPv4(addr).sin_addr.s_addr));
#ifdef	USE_IPV6
#ifdef	WORKAROUNDS
		case NET_ADDR_IPv6:
			return(local_v6);
		case NET_ADDR_UDP_IPv6:
			return(local_v6);
#endif	/* WORKAROUNDS */
#endif	/* USE_IPV6 */
		default:
			return(-1);
	}
}

#else /* end of SOLARIS */

#ifdef	linux

int dst_route_oif(net_addr *);
int dst_init();

int
unicast_init(void)
{
	if (dst_init(NULL) < 0) {
		init_failed = 1;
		NoUnicast = 1;
		log(LOG_INFO, 0, "Unicast routing information unavailable");
		return 0;
	}
	NoUnicast = 0;
	return 1;
}

/*
 * Find the kernel's idea of the route to destination DEST (in host order)
 */
int
unicast_route(net_addr *addr)
{
	return dst_route_oif(addr);
}

/* ------------------------------------------------------------------------ */
/*
 * libroute.c Interface to Linux routing engine.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include <linux/types.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>

#include <linux/rtnetlink.h>

#include "libnetlink.h"
#include "libroute.h"
#include "rsvp_daemon.h"

#define MAXDSTENTRIES 128
#define R_ASYNC_TMO	2
#define R_CANCEL_TMO	20
#define R_SYNC_TMO	5
#define R_GC_TMO	300
#define R_STALETIME	15
#define R_LOAD		30

#define MSG_TRUNC	0x20
#define MSG_DONTWAIT	0x40	/* Nonblocking io		 */

static struct rtnl_handle rt_nl;
static const __u32 zerokey[KEYLEN];
static struct dstentry *dsttable[256];
static struct dstentry *dstqueue;
static void (*dst_refresh)(struct dstentry*);
static int dsttable_size;
static int vif_by_if(int, int);

int dst_init(void (*refresh)(struct dstentry*))
{
	if (rtnl_open(&rt_nl, 0) < 0)
		return -1;
	dst_refresh = refresh;
	return rt_nl.fd;
}

static void dst_rmseq(struct dstentry *d)
{
	struct dstentry **dp;

	for (dp = &dstqueue; *dp; dp = &(*dp)->queue_next) {
		if (d == *dp) {
			*dp = d->queue_next;
			break;
		}
	}
	d->seq = 0;
}

static __inline__ void dst_addseq(struct dstentry *d)
{
	if (d->seq) {
		d->next = dstqueue;
		dstqueue = d;
	}
}

static struct dstentry * dst_findseq(__u32 seq)
{
	struct dstentry **dp, *d;

	for (dp = &dstqueue; (d=*dp)!=NULL; dp = &d->queue_next) {
		if (d->seq == seq) {
			*dp = d->queue_next;
			d->seq = 0;
			return d;
		}
	}
	return NULL;
}

static void release_path(struct dstentry *d)
{
	if (d->path) {
		free(d->path);
		d->path = NULL;
	}
}

static __inline__ unsigned hash(int family, const void *dst, const void *src)
{
	unsigned h;
	int offset = (family == AF_INET) ? 0 : 3;

	h = ((__u32*)dst)[offset];
	if (src)
		h ^= (((__u32*)src)[offset])>>4;
	h ^= h>>16;
	h ^= h>>8;
	return h&0xFF;
}

static __inline__ struct dstentry*
dst_lookup4(unsigned h, int iif, const void *dst, const void *src)
{
	__u32 saddr = src ? *(__u32*)src : 0;
	__u32 daddr = dst ? *(__u32*)dst : 0;
	struct dstentry *d;

	for (d = dsttable[h]; d; d = d->next) {
		if (d->family == AF_INET &&
		    d->iif == iif &&
		    d->dst[0] == daddr &&
		    d->src[0] == saddr)
			return d;
	}
	return NULL;
}


static __inline__ struct dstentry*
dst_lookup6(unsigned h, int iif, const void *dst, const void *src)
{
	struct dstentry *d;

	if (src == NULL)
		src = &zerokey;
	if (dst == NULL)
		dst = &zerokey;

	for (d = dsttable[h]; d; d = d->next) {
		if (d->family == AF_INET6 &&
		    d->iif == iif &&
		    memcmp(dst, d->dst, 16) == 0 &&
		    memcmp(dst, d->src, 16) == 0)
			return d;
	}
	return NULL;
}

static struct dstentry*
dst_lookup(int fam, int iif, const void *dst, const void *src)
{
	unsigned h = hash(fam, dst, src);

	return (fam==AF_INET?dst_lookup4:dst_lookup6)(h, iif, dst, src);
}

static void dst_gc()
{
	static int expires = R_GC_TMO;

	struct dstentry **dp, *d;
	time_t now = time(NULL);
	int h;

	for (h = 0; h<256; h++) {
		dp = &dsttable[h];
		while ((d=*dp)!=NULL) {
			if (!d->block) {
				if (now - d->accessed > expires) {
					*dp = d->next;
					if (d->seq)
						dst_rmseq(d);
					release_path(d);
					free(d);
					dsttable_size--;
					continue;
				}
				if (d->seq && now - d->updated > R_CANCEL_TMO)
					dst_rmseq(d);
			}
			dp = &d->next;
		}
	}
	if (dsttable_size > MAXDSTENTRIES)
		expires = ((expires+R_ASYNC_TMO)>>1) + 1;
	else
		expires = R_GC_TMO;
}

struct dstentry * dst_alloc(int family, void *dst, void *src, int iif)
{
	unsigned h;
	struct dstentry *d;

	d = calloc(1, sizeof(*d));
	if (d == NULL)
		return NULL;

	d->iif = iif;
	d->family = family;
	if (family == AF_INET) {
		d->dst[0] = *(__u32*)dst;
		if (src)
			d->src[0] = *(__u32*)src;
	}
#ifdef USE_IPV6
	else {
		memcpy(d->dst, dst, 16);
		memcpy(d->src, src ? : &zerokey, 16);
	}
#endif
	if (dsttable_size > MAXDSTENTRIES)
		dst_gc();

	dsttable_size++;
	h = hash(family,dst,src);
	d->next = dsttable[h];
	dsttable[h] = d;
	return d;
}



static int dst_request(struct dstentry *d)
{
	int rlen;
	char buf[256];
	struct nlmsghdr *n = (struct nlmsghdr *) buf;
	struct rtmsg *r = NLMSG_DATA(n);
	struct sockaddr_nl addr;

	n->nlmsg_type = RTM_GETROUTE;
	n->nlmsg_flags = NLM_F_REQUEST;
	n->nlmsg_len = NLMSG_LENGTH(sizeof(*r));
	n->nlmsg_pid = rt_nl.local.nl_pid;
	if (++rt_nl.seq == 0)
		rt_nl.seq = 1;
	n->nlmsg_seq = rt_nl.seq;

	if (d->seq)
		dst_rmseq(d);
	d->seq = n->nlmsg_seq;

	memset(r, 0, sizeof(*r));
	r->rtm_family = d->family;
	r->rtm_flags = RTM_F_NOTIFY;
	if (d->iif)
		addattr32(n, sizeof(buf), RTA_IIF, d->iif);

	if (d->family == AF_INET) {
		r->rtm_dst_len = 32;
		if (d->src[0])
			r->rtm_src_len = 32;
	}
#ifdef USE_IPV6
	else {
		r->rtm_dst_len = 128;
		if (d->src[0] || d->src[1] || d->src[2] || d->src[3])
			r->rtm_src_len = 128;
	}
#endif

	if (r->rtm_dst_len)
		addattr_l(n, sizeof(buf), RTA_DST, d->dst, r->rtm_dst_len/8);
	if (r->rtm_src_len)
		addattr_l(n, sizeof(buf), RTA_SRC, d->src, r->rtm_src_len/8);

	addr.nl_family = AF_NETLINK;
	addr.nl_groups = 0;
	addr.nl_pid = 0;

	if ((rlen = sendto(rt_nl.fd, buf, n->nlmsg_len, 0, (struct sockaddr *) &addr, sizeof(addr))) < 0) {
		d->seq = 0;
		return rlen;
	}

	dst_addseq(d);
	return 0;
}

static void netlink_accept(char *buf, int len)
{
	struct nlmsghdr *n = (void*)buf;
	struct rtmsg *r = NLMSG_DATA(n);
	struct rtattr *rta[RTA_MAX + 1];
	struct dstentry *d;
	time_t now = time(NULL);
	__u32 *dst = NULL;
	__u32 *src = NULL;
	int   iif = 0;
	int   err = 0;

	if (len < sizeof(struct nlmsghdr) || n->nlmsg_len > len)
		return;


	if (n->nlmsg_type == NLMSG_ERROR) {
		if (n->nlmsg_pid != rt_nl.local.nl_pid)
			return;
		err = ((struct nlmsgerr*)NLMSG_DATA(n))->error;
		d = dst_findseq(n->nlmsg_seq);
		if (d)
			goto failure;
		return;
	}

	if (n->nlmsg_type != RTM_NEWROUTE && n->nlmsg_type != RTM_DELROUTE)
		return;

	d = NULL;
	if (n->nlmsg_pid == rt_nl.local.nl_pid)
		d = dst_findseq(n->nlmsg_seq);

	if (parse_rtattr(rta, RTA_MAX, RTM_RTA(r), RTM_PAYLOAD(n))) {
		err = EINVAL;
		if (d)
			goto failure;
		return;
	}
	if (rta[RTA_DST])
		dst = RTA_DATA(rta[RTA_DST]);
	if (rta[RTA_SRC])
		src = RTA_DATA(rta[RTA_SRC]);
	if (rta[RTA_IIF])
		iif = *(int*)RTA_DATA(rta[RTA_IIF]);

	if (d == NULL || d->iif != iif) {
		struct dstentry *d1 = d;
		d = dst_lookup(r->rtm_family,iif,dst,src);
		if (d == NULL) {
			if (d1) {
				d = dst_alloc(r->rtm_family,dst,src,iif);
				if (d == NULL)
					return;
				d->state = R_INCOMPLETE;
			}
		}
		if (d->seq)
			dst_rmseq(d);
	}
	if (n->nlmsg_type == RTM_DELROUTE) {
		if (d->state&R_VALID)
			d->state = R_STALE;
		return;
	}
	if (r->rtm_type == RTN_LOCAL) {
		release_path(d);
		d->state = R_FRESH;
		d->error = 0;
	} else if (r->rtm_type == RTN_UNICAST || r->rtm_type == RTN_MULTICAST) {
		struct rtattr *path = rta[RTA_MULTIPATH];
		int do_free = 0;

		if (path == NULL) {
			struct rtnexthop *nh;
			int len = sizeof(struct rtnexthop);
			err = EINVAL;
			if (rta[RTA_OIF] == NULL)
				goto failure;
			if (rta[RTA_GATEWAY]) {
				if (memcmp(RTA_DATA(rta[RTA_GATEWAY]), dst,
					   RTA_PAYLOAD(rta[RTA_GATEWAY])))
					len += rta[RTA_GATEWAY]->rta_len;
				else
					rta[RTA_GATEWAY] = NULL;
			}
			path = malloc(RTA_LENGTH(len));
			if (path == NULL)
				return;
			path->rta_type = RTA_MULTIPATH;
			path->rta_len = RTA_LENGTH(len);
			nh = RTA_DATA(path);
			nh->rtnh_len = len;
			nh->rtnh_hops = 0;
			nh->rtnh_flags = r->rtm_flags;
			nh->rtnh_ifindex = *(int*)RTA_DATA(rta[RTA_OIF]);
			if (rta[RTA_GATEWAY])
				memcpy(RTNH_DATA(nh), rta[RTA_GATEWAY], rta[RTA_GATEWAY]->rta_len);
			do_free = 1;
		}
		if (d->path &&
		    d->path->rta_len == path->rta_len &&
		    memcmp(d->path, path, path->rta_len) == 0) {
			int old_state = d->state;
			d->state = R_FRESH;
			if (do_free)
				free(path);
			if (old_state&R_VALID)
				return;
			goto update;
		}
		if (!do_free) {
			struct rtattr *np = malloc(path->rta_len);
			if (np == NULL)
				return;
			memcpy(np, path, path->rta_len);
			path = np;
		}
		release_path(d);
		d->path = path;
		d->state = R_FRESH;
	} else {
		err = EINVAL;

failure:
		release_path(d);
		d->state = R_FAILED;
		d->error = err;
		d->updated = now;
		return;
	}

update:
	d->updated = now;

	if (!d->block && d->refresh) {
		d->refresh = 0;
		d->block++;
		if (dst_refresh)
			dst_refresh(d);
		d->block--;
	}
	return;
}

int netlink_read_wakeup()
{
	char buf[8192];
	int recvlen;
	struct sockaddr_nl nladdr;
	int addr_len = sizeof(nladdr);

	recvlen = recvfrom(rt_nl.fd, buf, sizeof(buf), MSG_DONTWAIT,
			   (struct sockaddr*)&nladdr, &addr_len);
	if (recvlen < 0) {
		if (errno == EAGAIN)
			return 1;
		return -1;
	}
	if (addr_len != sizeof(nladdr) || nladdr.nl_pid)
		return -1;
	netlink_accept(buf, recvlen);
	return 0;
}


struct dstentry *
dst_route(int family, void *dst, void *src, int iif, unsigned flags)
{
	time_t entered, now;
	struct dstentry *d;

	now = entered = time(NULL);

	d = dst_lookup(family, iif, dst, src);

	if (d == NULL) {
		if (flags&RTO_NOUPDATE)
			return NULL;

		d = dst_alloc(family, dst, src, iif);
		if (d == NULL)
			return NULL;
	}

	for (;;) {
		time_t age;
		int need_upd = 0;

		d->accessed = now;
		age = now - d->updated;
		if (d->state == R_FRESH) {
			if (age <= R_STALETIME)
				return d;
			d->state = R_STALE;
		}
		if ((d->state&(R_PROBE|R_INCOMPLETE)) && age > R_ASYNC_TMO) {
			d->state = R_FAILED;
			d->error = ETIMEDOUT;
		} else if (d->state == R_STALE) {
			need_upd = 1;
			d->state = R_PROBE;
		} else if (d->state == R_NONE ||
		    (d->state == R_FAILED && age > R_LOAD)) {
			need_upd = 1;
			d->updated = now;
			d->state = R_INCOMPLETE;
		}

		if (!(d->state&R_VALID))
			release_path(d);

		if (need_upd) {
			int err = dst_request(d);
			if (err) {
				d->state = R_FAILED;
				d->updated = now;
				d->error = errno;
				release_path(d);
			}
			d->block++;
			while (netlink_read_wakeup() == 0)
				/* NOTHING */;
			d->block--;
		}

		if (d->state&(R_VALID|R_FAILED))
			return d;

		if (flags&RTO_NOWAIT || now-entered > R_SYNC_TMO)
			return d;

		if (1) {
			fd_set fds;
			struct timeval tv;

			FD_ZERO(&fds);
			FD_SET(rt_nl.fd,&fds);
			tv.tv_sec = R_SYNC_TMO-(now-entered);
			tv.tv_usec = 0;
			select(rt_nl.fd+1, &fds, NULL, NULL, &tv);
			d->block++;
			netlink_read_wakeup();
			d->block--;
		}
		now = time(NULL);
	}
}


int dst_is_connected(struct dstentry *d)
{
	if (!(d->state&R_VALID))
		return 0;
	if (d->path==NULL ||
	    d->path->rta_len <= RTA_LENGTH(sizeof(struct rtnexthop)))
		return 1;
	return 0;
}

int dst_oif(struct dstentry *d)
{
	struct rtnexthop *nh;

	if (!(d->state&R_VALID))
		return 0;
	if (d->path==NULL)
		return 0;
	nh = RTA_DATA(d->path);
	return nh->rtnh_ifindex;
}

int dst_route_oif(net_addr *addr)
{
	int i;
	int family = AF_INET;
	net_addr *addr2;
	net_if *inf;
	struct dstentry *d;
	int best = -1;
	int out;
	int vif;

	addr2 = net_addr_ip(addr);

#ifdef USE_IPV6
	if (addr2->type == NET_ADDR_IPv6)
		family = AF_INET6;
#endif

	d = dst_route(family, &NET_GET_ADDR(addr2), NULL, 0, RTO_NOWAIT);
	if (d == NULL || !(d->state&R_VALID)) {
		d = NULL;
		goto fallback;
	}
	if (d->path == NULL)
		return -1;

	out = dst_oif(d);
	vif = vif_by_if(family, out);
	/* log(LOG_DEBUG, 0, "dst_route_oif: out = %d\n", out); */
	/* log(LOG_DEBUG, 0, "dst_route_oif: vif = %d\n", vif); */
	return vif;

fallback:
	/* First pass: check primary addresses */

	for (i=0; i<if_num; i++) {
		if (IsNumAPI(i))
			continue;
		if (IF_DOWN(i))
		        continue;
		inf = &GET_IF(i);
		if (NET_GET_TYPE(inf) != NET_IF_PHY)
			continue;
		if (net_addr_equal(addr2,&NET_GET_IF_PHY_ADDR(inf)))
			return i;
	}

	if (best >= 0)
		return best;

	if (d) {
		best = local_v4;
#ifdef USE_IPV6
		if (family == AF_INET6)
			best = local_v6;
#endif
	}		
	return best;
}

static int vif_by_if(int family, int ifindex)
{
	int i;
	net_if *inf2;
	int type = NET_ADDR_IPv4;

#ifdef USE_IPV6
	if (family == AF_INET6)
		type = NET_ADDR_IPv6;
#endif

	for (i=0; i<if_num; i++) {
		if (IsNumAPI(i))
			continue;
		inf2 = &GET_IF(i);
		if (NET_GET_TYPE(inf2) == NET_IF_PHY &&
		    NET_GET_IF_PHY_ID(inf2) == ifindex &&
		    NET_GET_IF_PHY_ADDR(inf2).type == type)
			return i;
	}
	log(LOG_WARNING, 0, "ifindex %d/%d is not mapped to a vif\n", ifindex, type);
	return -1;
}

#elif	sgi_53

/* Do not have support for access to unicast routing table in IRIX 5.3.
 */
int
unicast_init()
	{
#ifdef HOST_ONLY
	return(0);
#else
	return(1);
#endif
}

int
unicast_route1(u_long dest)
{
	/*  This dummy routine is here to allow pre-routing-socket code to
	 *  compile.  It should never be called.
	 */
	assert(0);
	return(0);
}

#else  /* !sgi_53 && !linux */

/*	Following code should work for many systems derived from
 *	older BSDs, but it is known to work for Sun OS.
 */

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */
#include <fcntl.h>
#include <kvm.h>
#include <nlist.h>
#include <stdio.h>

#include <sys/sockio.h>
#include <sys/mbuf.h>

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>

#include <netdb.h>

#include <assert.h>

#include <syslog.h>

struct nlist nl[] = {
#define	N_RTHOST	0
	{ "_rthost" },
#define	N_RTNET		1
	{ "_rtnet" },
#define N_RTHASHSIZE	2
	{ "_rthashsize" },
	{""}
};

kvm_t	*kd;
char	usage[] = "[-n] host";

struct xroute {
	char ifname[IFNAMSIZ];
	struct sockaddr_in netmask;
	struct sockaddr_in out_if;
	struct sockaddr_in dst;
	struct sockaddr_in gw;
} routes[256];

int nhost = 0;
int nroute = 0;

struct xroute default_route;
int have_default = 0;

int unicast_fd;

/* forward declarations */
int read_routes();
int net_match(struct sockaddr_in *, struct xroute *);
int rtfindx(struct sockaddr_in *);
int ifname_to_if(char *);
int kread(unsigned long addr, char *buf, unsigned nbytes);
static int inet_netmatch(struct sockaddr_in *, struct sockaddr_in *);

int
unicast_init() {
	if ((kd = kvm_open(NULL, NULL, NULL, O_RDONLY, NULL)) == NULL) {
		log(LOG_INFO, 0, "Unicast routing information unavailable\n");
		init_failed = 1;
		NoUnicast = 1;
		return(0);
	}
	if (kvm_nlist(kd, nl) < 0) {
		log(LOG_ERR, 0, "netstat: bad namelist\n");
		init_failed = 1;
		NoUnicast = 1;
		return(0);
	}
	unicast_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (unicast_fd < 0) {
		log(LOG_ERR, errno, "unicast socket");
		init_failed = 1;
		NoUnicast = 1;
		return(0);
	}
	NoUnicast = 0;
	return(1);
}

int
unicast_route1(u_long addr) {
	int i;
	struct sockaddr_in sin;
	if (init_failed)
		return(-1);
	if (read_routes() == -1)
		return(-1);
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = addr;
	i = rtfindx(&sin);
	if (i == -1 && have_default) {
		if (IsDebug(DEBUG_ROUTE)) {
			log(LOG_DEBUG, 0, "Using default route ...\n");
			log(LOG_DEBUG, 0, "route out interface %s\n", 
						default_route.ifname);
		}
		return(ifname_to_if(default_route.ifname));
	} else if (i == -1) {
		if (IsDebug(DEBUG_ROUTE))
			log(LOG_DEBUG, 0, "No route\n");
		return(-1);
	}
	if (IsDebug(DEBUG_ROUTE))
		log(LOG_DEBUG, 0, "route out interface %s\n", routes[i].ifname);
	return(ifname_to_if(routes[i].ifname));
}

int
unicast_route(net_addr *addr)
{
	switch (NET_GET_TYPE(addr)) {
		case NET_ADDR_IPv4:
			return(unicast_route1(NET_GET_ADDR_IPv4(addr).s_addr));
		case NET_ADDR_UDP_IPv4:
			return(unicast_route1(NET_GET_ADDR_UDP_IPv4(addr)
				.sin_addr.s_addr));
#ifdef	USE_IPV6
#ifdef	WORKAROUNDS
		case NET_ADDR_IPv6:
			return(local_v6);
		case NET_ADDR_UDP_IPv6:
			return(local_v6);
#endif	/* WORKAROUNDS */
#endif	/* USE_IPV6 */
		default:
			return(-1);
	}
}

int
ifname_to_if(char *name) {
	int i;
	for (i = 0; i < if_num; i++)
		if (strncmp(if_vec[i].if_name, name, IFNAMSIZ) == 0)
			return(i);
	log(LOG_ERR, 0, "Couldn't find interface %s\n", name);
	return(-1);
}
int
read_routes() {
	struct mbuf mbf, *mbfp = &mbf;
	register struct mbuf *m;
	struct mbuf **rhash;
	int i, doinghost = 1;
	int hashsize;
	struct ifreq ifrbuf, *ifr = &ifrbuf;
	int status;

	kread(nl[N_RTHASHSIZE].n_value, (char *)&hashsize, sizeof (hashsize));
	rhash = (struct mbuf **)malloc( hashsize*sizeof (struct mbuf *) );
	kread(nl[N_RTHOST].n_value, (char *)rhash, hashsize*sizeof (struct mbuf *));
	nroute = 0;
again:
	for (i = 0; i < hashsize; i++) {
		if (rhash[i] == 0)
			continue;
		m = rhash[i];
		while (m) {
			struct ifnet ifz, *ifp = &ifz;
			struct rtentry *rt;
			struct ifaddr ifx, *ifxp = &ifx;
			char ifbuf[IFNAMSIZ];

			kread((u_long)m, (char *)mbfp, sizeof(struct mbuf));
			rt = mtod(mbfp, struct rtentry *);
			assert(rt->rt_ifp);
			routes[nroute].dst = *(struct sockaddr_in *)&rt->rt_dst;
			routes[nroute].gw = *(struct sockaddr_in *)&rt->rt_gateway;
			kread((u_long)rt->rt_ifp, (char *)ifp, sizeof(struct ifnet));
			assert(ifp->if_addrlist);
			kread((u_long)ifp->if_addrlist, (char *)ifxp, sizeof(struct ifaddr));
			assert(ifxp->ifa_addr.sa_family == AF_INET);
			routes[nroute].out_if = *(struct sockaddr_in *)&ifx.ifa_addr;
			kread((u_long)ifp->if_name, ifbuf, IFNAMSIZ);
			sprintf(routes[nroute].ifname, "%s%d", ifbuf, 
				ifp->if_unit);
			memcpy(ifr->ifr_name, routes[nroute].ifname, IFNAMSIZ);
			status = ioctl(unicast_fd, SIOCGIFNETMASK, ifr);
			if (status < 0) {
				perror("ioctl");
				return(-1);
			}
			routes[nroute].netmask = *(struct sockaddr_in *)&ifr->ifr_addr;
			if (routes[nroute].dst.sin_addr.s_addr == INADDR_ANY) {
				default_route = routes[nroute];
				have_default = 1;
			} else
				nroute++;
			assert(nroute < 256);
			m = mbfp->m_next;
		}
	}
	if (doinghost) {
		kread(nl[N_RTNET].n_value, (char *)rhash, hashsize*sizeof (struct mbuf *));
		nhost = nroute;
		doinghost = 0;
		goto again;
	}
	free(rhash);
	return(1);
}

int
kread(unsigned long addr, char *buf, unsigned nbytes) {
	return kvm_read(kd, addr, buf, nbytes);
}

/*
 * Find a route to dst as the kernel would.
 */
int
rtfindx(struct sockaddr_in *dst) {
	int i;

	for (i = 0; i < nhost; i++)
		if (memcmp(&routes[i].dst, dst, sizeof(struct sockaddr)) == 0)
			return(i);
	for (i = nhost; i < nroute; i++)
		if (net_match(dst, &routes[i]))
			return(i);
	return(-1);
}

int
net_match(struct sockaddr_in *dst, struct xroute *rp) {
	u_long nm = rp->netmask.sin_addr.s_addr;
	if (inet_netmatch(dst, &rp->dst)) {
		u_long ip1 = dst->sin_addr.s_addr;
		u_long ip2 = rp->dst.sin_addr.s_addr;
		if (inet_netmatch(dst, &rp->out_if)) {
			if ((ip1 & nm) == ip2)
				return(1);
			else
				return(0);
		} else
			return(1);
	}
	return(0);
}

int
inet_netmatch(struct sockaddr_in *sin1, struct sockaddr_in *sin2) {
	return (inet_netof(sin1->sin_addr) == inet_netof(sin2->sin_addr));
}
#endif /* ! sgi */
#endif /* ! SOLARIS */
#endif /* ! PF_ROUTE */

#ifdef SOLARIS
/*
 * Copyright (c) Sun Microsystems, Inc.  1994. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative works
 * for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or
 * derivative work. Sun Microsystems grants no other licenses expressed or
 * implied. The Sun Microsystems  trade name should not be used in any
 * advertising without its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */
/* Portions of this code were dereived from Solaris route.c and netstat.c */

#include <fcntl.h>
#include <sys/stropts.h>
#include <sys/socket.h>
#include <net/route.h>
#include <netinet/in.h>
#include <inet/ip.h>
#include <sys/tihdr.h>
#include <sys/tiuser.h>
#include <inet/mib2.h>

/*
 * Following Solaris code was supplied by Don Hoffman of Sun Microsystems.
 */

/*
 * Defines
 */
#define MAX_ROUTES	(4 * 512)
#ifndef T_CURRENT
#define T_CURRENT       MI_T_CURRENT
#endif

#define SOLARIS_2_6	6

static int t_current = T_CURRENT;

/*
 * Structure definitions
 */
typedef struct mib_item_s {
	struct mib_item_s *next_item;
	long group;
	long mib_id;
	long length;
	char *valp;
} mib_item_t;

struct xroute {
	char ifname[IFNAMSIZ];
	struct sockaddr_in netmask;
	struct sockaddr_in out_if;
	struct sockaddr_in dst;
	struct sockaddr_in gw;
};

/*
 * Forward function declarations
 */
static mib_item_t *mibget(int sd);
static void mibfree(mib_item_t *);
static int is_host(mib2_ipRouteEntry_t * rp);
static int is_gateway(mib2_ipRouteEntry_t * rp);
int putmsg(int, struct strbuf *, struct strbuf *, int);
int getmsg(int, struct strbuf *, struct strbuf *, int *);

/*
 * Globals
 */
static int sd;
static struct xroute routes[MAX_ROUTES];
static struct xroute default_route;
static int have_default = 0;
static int nhost = 0;
static int nroute = 0;

static int
ifname_to_if(char *name)
{
	int i;

	if (name)
		fprintf(stderr, "ifname is %s\n", name);

	for (i = 0; i < if_num; i++)
		if (strncmp(if_vec[i].if_name, name, IFNAMSIZ) == 0)
			return (i);
	log(LOG_ERR, 0, "Couldn't find interface %s\n", name);
#ifdef SUNMOD
	return (0);
#else
	return (-1);
#endif
}

#ifdef SBM
u_long
unicast_route_gw_specific(u_long addr)
{
	int i;
	struct sockaddr_in sin;

	if (read_routes() == -1)
		return (-1);

	memset(&sin, 0, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = addr;
	i = rtfindx(&sin);

	if (i == -1 && have_default) {
		if (IsDebug(DEBUG_ROUTE)) {
			log(LOG_DEBUG, 0, "Using default route ...\n");
			log(LOG_DEBUG, 0, "route out interface %s\n",
				default_route.ifname);
		}
		return (default_route.gw.sin_addr.s_addr);
	} else if (i >= 0)
		return (routes[i].gw.sin_addr.s_addr);
	else
		return -1;
}

#endif

static int
inet_netmatch(struct sockaddr_in * sin1, struct sockaddr_in * sin2)
{
	return (inet_netof(sin1->sin_addr) == inet_netof(sin2->sin_addr));
}

static int
net_match(struct sockaddr_in * dst, struct xroute * rp)
{
	u_long nm = rp->netmask.sin_addr.s_addr;

	if (inet_netmatch(dst, &rp->dst)) {
		u_long ip1 = dst->sin_addr.s_addr;
		u_long ip2 = rp->dst.sin_addr.s_addr;

		if (inet_netmatch(dst, &rp->out_if)) {
			if ((ip1 & nm) == ip2)
				return (1);
			else
				return (0);
		} else
			return (1);
	}
	return (0);
}

/*
 * Find a route to dst as the kernel would.
 */
static int
rtfindx(struct sockaddr_in *dst)
{
	struct sockaddr_in	*sockp;
	int 			i;

	for (i = 0; i < nhost; i++) {
		sockp = &(routes[i].dst);
#ifdef notdef
		log(LOG_DEBUG, 0, "routes[%d].dst, dst is %x,%x\n", 
			i, sockp->sin_addr.s_addr, dst->sin_addr.s_addr); 
#endif
		if (memcmp(sockp, dst, sizeof (struct sockaddr)) == 0)
			return (i); 
	} 

	for (i = nhost; i < nroute; i++)
		if (net_match(dst, &routes[i]))
			return (i);
	return (-1);
}

static int
read_routes()
{
	mib_item_t *item;
	mib_item_t *first_item = nilp(mib_item_t);
	mib2_ipRouteEntry_t *rp;
	int ret_code = 1;
	int doinghost;
	int i;


	/* Get the route list. */
	/*
	 * TBD: don't need to do this every time.  Only to a new mibget every X
	 * seconds?
	 */
	if ((first_item = mibget(sd)) == nilp(mib_item_t)) {
		close(sd);
		ret_code = -1;
		goto leave;
	}
	/* Look up the entry */
	have_default = 0;
	nroute = 0;
	nhost = 0;

	/* host routes first */
	doinghost = 1;
again:
	for (item = first_item; item; item = item->next_item) {
		/* skip all the other trash that comes up the mib stream */
		if ((item->group != MIB2_IP) || (item->mib_id != MIB2_IP_21))
			continue;

		rp = (mib2_ipRouteEntry_t *) item->valp;
		for (; (u_long) rp < (u_long)(item->valp + item->length);
				rp++) {
			/* skip routes that aren't what we want this pass */
			if (doinghost) {
				if (!is_host(rp))
					continue;
			} else {
				if (!is_gateway(rp))
					continue;
			}

			/* fill in the blanks */
			memset(&routes[nroute], 0, sizeof (struct xroute));
			routes[nroute].dst.sin_family = AF_INET;
			routes[nroute].dst.sin_addr.s_addr = rp->ipRouteDest;
			routes[nroute].gw.sin_family = AF_INET;
			routes[nroute].gw.sin_addr.s_addr = rp->ipRouteNextHop;
			routes[nroute].netmask.sin_family = AF_INET;
			routes[nroute].netmask.sin_addr.s_addr =
				rp->ipRouteMask;
			routes[nroute].out_if.sin_family = AF_INET;
			routes[nroute].out_if.sin_addr.s_addr =
				rp->ipRouteInfo.re_src_addr;

			if (rp->ipRouteIfIndex.o_length >= IFNAMSIZ)
				rp->ipRouteIfIndex.o_length = IFNAMSIZ - 1;
			for (i = 0; i < rp->ipRouteIfIndex.o_length; i++)
				routes[nroute].ifname[i] =
					rp->ipRouteIfIndex.o_bytes[i];
			routes[nroute].ifname[i] = '\0';

			if (routes[nroute].dst.sin_addr.s_addr == INADDR_ANY) {
				default_route = routes[nroute];
				have_default = 1;
			} else
				nroute++;

		}
	}

	/* net routes next */
	if (doinghost) {
		nhost = nroute;
		doinghost = 0;
		goto again;
	}
leave:
	mibfree(first_item);
	first_item = nilp(mib_item_t);
	return ret_code;
}

static int
is_host(mib2_ipRouteEntry_t * rp)
{
	if (rp->ipRouteMask == (IpAddress) - 1)
		return 1;
	else
		return 0;
}

static int
is_gateway(mib2_ipRouteEntry_t * rp)
{
#ifndef  PF_ROUTE
	if (rp->ipRouteInfo.re_ire_type == IRE_GATEWAY ||
		rp->ipRouteInfo.re_ire_type == IRE_NET ||
		rp->ipRouteInfo.re_ire_type == IRE_ROUTE_ASSOC ||
		rp->ipRouteInfo.re_ire_type == IRE_ROUTE_REDIRECT)
		return 1;
	else
		return 0;
#else
	if (rp->ipRouteInfo.re_ire_type == IRE_DEFAULT ||
		rp->ipRouteInfo.re_ire_type == IRE_PREFIX ||
		rp->ipRouteInfo.re_ire_type == IRE_HOST ||
		rp->ipRouteInfo.re_ire_type == IRE_HOST_REDIRECT)
		return 1;
	else
		return 0;
#endif
}

static mib_item_t * 
mibget (int sd)
{
	char			buf[512];
	int			flags;
	int			i, j, getcode;
	struct strbuf		ctlbuf, databuf;
	struct T_optmgmt_req	*tor = (struct T_optmgmt_req *)buf;
	struct T_optmgmt_ack	*toa = (struct T_optmgmt_ack *)buf;
	struct T_error_ack	*tea = (struct T_error_ack *)buf;
	struct opthdr		*req;
	mib_item_t		*first_item = nilp(mib_item_t);
	mib_item_t		*last_item  = nilp(mib_item_t);
	mib_item_t		*temp;
	int debug = 0;
	
	tor->PRIM_type = T_OPTMGMT_REQ;
	tor->OPT_offset = sizeof(struct T_optmgmt_req);
	tor->OPT_length = sizeof(struct opthdr);
	tor->MGMT_flags = t_current;
	req = (struct opthdr *)&tor[1];
	req->level = MIB2_IP;		/* any MIB2_xxx value ok here */
	req->name  = 0;
	req->len   = 0;

	ctlbuf.buf = buf;
	ctlbuf.len = tor->OPT_length + tor->OPT_offset;
	flags = 0;
	if (putmsg(sd, &ctlbuf, nilp(struct strbuf), flags) == -1) {
		perror("mibget: putmsg(ctl) failed");
		goto error_exit;
	}
	/*
	 * each reply consists of a ctl part for one fixed structure
	 * or table, as defined in mib2.h.  The format is a T_OPTMGMT_ACK,
	 * containing an opthdr structure.  level/name identify the entry,
	 * len is the size of the data part of the message.
	 */
	req = (struct opthdr *)&toa[1];
	ctlbuf.maxlen = sizeof(buf);
	for (j=1; ; j++) {
		flags = 0;
		getcode = getmsg(sd, &ctlbuf, nilp(struct strbuf), &flags);
		if (getcode == -1) {
			perror("mibget getmsg(ctl) failed");
			if (debug) {
				fprintf(stderr, "#   level   name    len\n");
				i = 0;
				for (last_item = first_item; last_item; 
					last_item = last_item->next_item)
					printf("%d  %4ld   %5ld   %ld\n", ++i,
						last_item->group, 
						last_item->mib_id, 
						last_item->length);
			}
			goto error_exit;
		}
		if (getcode == 0
		&& ctlbuf.len >= sizeof(struct T_optmgmt_ack)
		&& toa->PRIM_type == T_OPTMGMT_ACK
		&& toa->MGMT_flags == T_SUCCESS
		&& req->len == 0) {
			if (debug)
				printf("mibget getmsg() %d returned EOD (level %ld, name %ld)\n", 
				       j, req->level, req->name);
			return first_item;		/* this is EOD msg */
		}

		if (ctlbuf.len >= sizeof(struct T_error_ack)
		&& tea->PRIM_type == T_ERROR_ACK) {
			fprintf(stderr, 
			"mibget %d gives (%d) T_ERROR_ACK: TLI_error = 0x%lx, UNIX_error = 0x%lx\n",
				j, getcode, tea->TLI_error, tea->UNIX_error);
			errno = (tea->TLI_error == TSYSERR)
				? tea->UNIX_error : EPROTO;
			goto error_exit;
		}
			
		if (getcode != MOREDATA
		|| ctlbuf.len < sizeof(struct T_optmgmt_ack)
		|| toa->PRIM_type != T_OPTMGMT_ACK
		|| toa->MGMT_flags != T_SUCCESS) {
			printf(
			"mibget getmsg(ctl) %d returned %d, ctlbuf.len = %d, PRIM_type = %ld\n",
				 j, getcode, ctlbuf.len, toa->PRIM_type);
			if (toa->PRIM_type == T_OPTMGMT_ACK)
				printf(
				"T_OPTMGMT_ACK: MGMT_flags = 0x%lx, req->len = %ld\n", 
					toa->MGMT_flags, req->len);
			errno = ENOMSG;
			goto error_exit;
		}

		temp = (mib_item_t *)malloc(sizeof(mib_item_t));
		if (!temp) {
			perror("mibget malloc failed");
			goto error_exit;
		}
		if (last_item)
			last_item->next_item = temp;
		else
			first_item = temp;
		last_item = temp;
		last_item->next_item = nilp(mib_item_t);
		last_item->group = req->level;
		last_item->mib_id = req->name;
		last_item->length = req->len;
		last_item->valp = (char *)malloc(req->len);
		if (debug)
			printf(
			"msg %d:  group = %4ld   mib_id = %5ld   length = %ld\n", 
				j, last_item->group, last_item->mib_id, 
				last_item->length);

		databuf.maxlen = last_item->length;
		databuf.buf    = last_item->valp;
		databuf.len    = 0;
		flags = 0;
		getcode = getmsg(sd, nilp(struct strbuf), &databuf, &flags);
		if (getcode == -1) {
			perror("mibget getmsg(data) failed");
			goto error_exit;
		} else if (getcode != 0) {
			printf(
			"mibget getmsg(data) returned %d, databuf.maxlen = %d, databuf.len = %d\n",
				 getcode, databuf.maxlen, databuf.len);
			goto error_exit;
		}
	}

error_exit:;
	while (first_item) {
		last_item = first_item;
		first_item = first_item->next_item;
		free(last_item);
	}
	return first_item;
}


static void
mibfree(mib_item_t * first_item)
{
	mib_item_t *last_item;

	while (first_item) {
		last_item = first_item;
		first_item = first_item->next_item;
		free(last_item);
	}
}

int
unicast_init_specific(void)
{
	int	release;

	release = getSolarisRelease();
	if (release < SOLARIS_2_6) {
		t_current = 0x100;
	}

	sd = open("/dev/ip", O_RDWR);
	if (sd == -1) {
		log(LOG_INFO, 0, "Unicast routing information unavailable\n");
		init_failed = 1;
		NoUnicast = 1;
		return 0;
	}
	return (1);
}

int
unicast_route_specific(u_long addr)
{
	int i;
	struct sockaddr_in sin;

	if (init_failed)
		return (-1);
	if (read_routes() == -1)
		return (-1);

	memset(&sin, 0, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = addr;

	i = rtfindx(&sin);
	if (i == -1 && have_default) {
		if (strlen(default_route.ifname) == 0) {
			/*
			 * The interface name is not set on def route.  Look it
			 * up.
			 */
			if ((i = rtfindx(&default_route.gw)) == -1)
				return (-1);
		}
		if (IsDebug(DEBUG_ROUTE)) {
			log(LOG_DEBUG, 0, "Using default route ...\n");
			log(LOG_DEBUG, 0, "route out interface %s\n",
				default_route.ifname);
		}
		return (ifname_to_if(default_route.ifname));
	} else if (i == -1) {
		if (IsDebug(DEBUG_ROUTE))
			log(LOG_DEBUG, 0, "No route\n");
		return (-1);
	}
	if (strlen(routes[i].ifname) == 0) {
		/*
		 * The interface name is not set on GW routes.  Look it up.
		 */
		if ((i = rtfindx(&routes[i].gw)) == -1)
			return (-1);
	}
	if (IsDebug(DEBUG_ROUTE))
		log(LOG_DEBUG, 0, "route out interface %s\n", routes[i].ifname);
	return (ifname_to_if(routes[i].ifname));
}

#endif
