/*
 * @(#) $Id: rsvp_socks.c,v 1.1.1.1 2000/05/08 22:51:24 wenqing Exp $
 */

/************************ rsvp_socks.h *******************************
 *                                                                   *
 *    IPv4/IPv6 Sockets Interface based on RFC2133 and more          *
 *                                                                   *
 *********************************************************************/

/****************************************************************************

            RSVPD -- ReSerVation Protocol Daemon

                USC Information Sciences Institute
                Marina del Rey, California

		Original Version: Shai Herzog, Nov. 1993.
		Current Version:  Bob Lindell, July 1997.

  Copyright (c) 1999 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies, and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#ifdef	linux
#include <linux/sockios.h>
#include <linux/types.h>
#include <linux/rtnetlink.h>
#include <sys/uio.h>
#include <errno.h>
#include "libnetlink.h"
#else	/* linux */
#include <sys/sockio.h>
#endif	/* linux */
#include "rsvp_socks.h"
#ifdef	__FreeBSD__
#include <sys/sysctl.h>
#include <net/if_dl.h>
#include <net/route.h>
#ifdef	USE_IPV6
#ifdef IPV6_INRIA_VERSION
#include <netinet/in6_var.h>
#endif
#endif	/* USE_IPV6 */
#endif	/* __FreeBSD__ */

#ifdef	sun
extern int gethostname(char *,int);
#if	(!defined(__SVR4) && !defined(__svr4__))
extern int ioctl(int,int,caddr_t);
extern int socket(int,int,int);
#endif	/* (!defined(__SVR4) && !defined(__svr4__)) */
#endif	/* sun */

#define	PREFIX(x)	prefix((char *) &(x),NBITS(x))

#ifdef	USE_IPV6
#define	_AF_INET	AF_INET6
#ifdef	sun
#define	IFREQ		v6conf
#define	IF_ADDR		v6cf_addr_in
#define	IF_NAME		v6cf_name
#define	IFREQMASK	v6maskreq
#define	IF_MASKNAME	v6mr_name
#define	IF_MASKADDR	v6mr_mask
#define	IFREQADDR	v6addrreq
#define	IF_ADDRNAME	v6ar_name
#define	IF_ADDRADDR	v6ar_addr
#define	PREFIX6(x)	x
#define	_SIOCGIFCONF	SIOCGIFV6CONF
#define	_SIOCGIFADDR	SIOCGIFV6ADDR
#define	SIOCGIFNETMASK6	SIOCGIFV6MASK
#endif	/* sun */
#else	/* USE_IPV6 */
#define	_AF_INET	AF_INET
#define	IFREQ		ifreq
#define	IF_ADDR		ifr_addr
#define	IF_NAME		ifr_name
#define	IFREQADDR	IFREQ
#define	IF_ADDRNAME	IF_NAME
#define	IF_ADDRADDR	IF_ADDR
#define	_SIOCGIFCONF	SIOCGIFCONF
#define	_SIOCGIFADDR	SIOCGIFADDR
#endif	/* USE_IPV6 */

#ifndef	__FreeBSD__
static int if_indextoprefix(unsigned int n);
#endif	/* __FreeBSD__ */
static unsigned int if_indextoflags(unsigned int n);
static unsigned int if_indextomtu(unsigned int n);
#ifndef	linux
static unsigned int prefix(unsigned char *p,unsigned int n);
#endif	/* linux */

#ifdef	sun
const struct in_addr inaddr_any = {{{INADDR_ANY}}};
#else	/* sun */
const struct in_addr inaddr_any = {INADDR_ANY};
#endif	/* sun */

#if	(defined(__FreeBSD__) || defined(sun))
static struct if_nameindex ifs[MAX_INTERFACES];
#endif	/* (defined(__FreeBSD__) || defined(sun)) */

#ifdef	__FreeBSD__

/*
 * Expand the compacted form of addresses as returned via the
 * configuration read via sysctl().
 */

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n) (x += ROUNDUP((n)->sa_len))

/*
 * FreeBSD and other *BSD use sysctl to get the interface list.
 * This code was taken from ifconfig.8.  The code was contributed
 * to rsvpd from Kenjiro Cho <kjc@csl.sony.co.jp>.
 */

static struct if_attributes ifap[MAX_INTERFACES][MAX_INTERFACES];
static struct sockaddr *prefixes[MAX_INTERFACES][MAX_INTERFACES];

struct if_nameindex *
if_nameindex()
{
    struct	if_msghdr *ifm;
    struct	ifa_msghdr *ifam;
    struct	sockaddr *s;
    struct	sockaddr_dl *sdl;
    char	*buf, *lim, *next, *p, *cp;
    size_t	needed;
    int 	mib[6], index, i, j;

    for (i = 0; i < MAX_INTERFACES; i++) {
	for (j = 0; ifap[i][j].addr != NULL; j++) {
		free(ifap[i][j].addr);
		ifap[i][j].addr = NULL;
	}
	for (j = 0; prefixes[i][j] != 0; j++) {
		free(prefixes[i][j]);
		prefixes[i][j] = NULL;
	}
	if (ifs[i].if_name != NULL)
		if (strcmp(ifs[i].if_name,"") != 0)
			free(ifs[i].if_name);
	ifs[i].if_name = NULL;
    }

    mib[0] = CTL_NET;
    mib[1] = PF_ROUTE;
    mib[2] = 0;
#ifdef	USE_IPV6
    mib[3] = 0;		/* address family */
#else	/* USE_IPV6 */
    mib[3] = AF_INET;	/* address family */
#endif	/* USE_IPV6 */
    mib[4] = NET_RT_IFLIST;
    mib[5] = 0;

    if (FAILED(sysctl(mib, 6, NULL, &needed, NULL, 0)))
	return(NULL);
    if ((buf = malloc(needed)) == NULL)
	return(NULL);
    if (FAILED(sysctl(mib, 6, buf, &needed, NULL, 0))) {
	free(buf);
	return(NULL);
    }
    lim = buf + needed;

    for (next = buf; next < lim; next += ifm->ifm_msglen) {

	ifm = (struct if_msghdr *)next;

	switch (ifm->ifm_type) {
	case RTM_IFINFO:
	    index = ifm->ifm_index - 1;
	    if (index < 0 || index >= MAX_INTERFACES)
		continue;

	    ifs[index].if_index = index + 1;  /* real ifm_index value */
	    sdl = (struct sockaddr_dl *)(ifm + 1);
	    if ((ifs[index].if_name = malloc(sdl->sdl_nlen+1)) == NULL)
		break;
	    strncpy(ifs[index].if_name, sdl->sdl_data, sdl->sdl_nlen);
	    ifs[index].if_name[sdl->sdl_nlen] = '\0';
	    break;

	case RTM_NEWADDR:
	    ifam = (struct ifa_msghdr *)ifm;
	    index = ifam->ifam_index - 1;
	    if (index < 0 || index >= MAX_INTERFACES)
		continue;
	    p = (char *)(ifam + 1);
	    for (i = 1; i < 0x100; i <<= 1) {
		if ((ifam->ifam_addrs & i) == 0)
		    continue;
		if (i == RTA_IFA) {
		    s = (struct sockaddr *)p;
		    switch (s->sa_family) {
		    case AF_INET:
#ifdef	USE_IPV6
		    case AF_INET6:
#endif	/* USE_IPV6 */
			if ((cp = malloc(s->sa_len)) == NULL)
			    break;
			memcpy(cp, s, s->sa_len);
			for (j = 0; ifap[index][j].addr != NULL; j++)
			    ;
			ifap[index][j++].addr = (struct sockaddr *) cp;
			ifap[index][j].addr = NULL;
			break;
		    default:
			break;
		    }
		}
		if (i == RTA_NETMASK) {
		    s = (struct sockaddr *)p;
		    if ((cp = malloc(s->sa_len)) == NULL)
			 break;
		    memcpy(cp, s, s->sa_len);
		    for (j = 0; prefixes[index][j] != 0; j++)
			;
		    prefixes[index][j] = (struct sockaddr * ) cp;
		}
		ADVANCE(p, (struct sockaddr *)p);
	    }
	    break;

	default:
	    break;

	}
    }
    free(buf);

    /*
     * make sure there's no index gap, other routines use NULL-if_name
     * as an end mark.
     */
    for (index = MAX_INTERFACES -1; index > 0; index--)
	if (ifs[index].if_name != NULL)
	    break;
    for (; index > 0; index--)
	if (ifs[index].if_name == NULL)
	    ifs[index].if_name = "";

    return(ifs);
}

#else	/* __FreeBSD__ */

#ifdef	linux

/*
 *	The code was borrowed from the linux port of rsvpd done by Alexey
 *	Kuznetsov <kuznet@ms2.inr.ac.ru>
 */

static struct if_attributes ifap[MAX_INTERFACES];

int store_addrinfo(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	int i;
	struct ifaddrmsg *ifa = NLMSG_DATA(n);
	struct rtattr *rta = IFA_RTA(ifa);
	int len = n->nlmsg_len;
	struct rtattr * rta_tb[IFA_MAX+1];
	struct in_addr in;
#ifdef USE_IPV6
	struct in6_addr in6;
#endif	/* USE_IPV6 */

	if (n->nlmsg_type != RTM_NEWADDR)
		return 0;
	len -= NLMSG_LENGTH(sizeof(*ifa));
	if (len < 0) {
		errno = 0;
		return -1;
	}

	if (ifa->ifa_flags&IFA_F_TENTATIVE)
		return 0;

	parse_rtattr(rta_tb, IFA_MAX, rta, len);

	i = ifa->ifa_index;
	if (i >= MAX_INTERFACES)
		return 0;

	ifap[i].prefix = ifa->ifa_prefixlen;

	switch (ifa->ifa_family) {
		case AF_INET:
			memcpy(&in,RTA_DATA(rta_tb[IFA_ADDRESS]),
				sizeof(in));
			ifap[i].addr = malloc(sizeof(struct sockaddr_in));
			NET_SOCKADDR_IPv4((struct sockaddr_in *)
				ifap[i].addr,in);
			break;
#ifdef USE_IPV6
		case AF_INET6:
			memcpy(&in6,RTA_DATA(rta_tb[IFA_ADDRESS]),
				sizeof(in6));
			ifap[i].addr = malloc(sizeof(struct sockaddr_in6));
			NET_SOCKADDR_IPv6((struct sockaddr_in6 *)
				ifap[i].addr,in6);
			break;
#endif	/* USE_IPV6 */
	}
/*
	if (rta_tb[IFA_LOCAL] == NULL) {
		rta_tb[IFA_LOCAL] = rta_tb[IFA_ADDRESS];
		if (rta_tb[IFA_LOCAL] == NULL)
			return 0;
	}
	if (rta_tb[IFA_ADDRESS] == NULL)
		rta_tb[IFA_ADDRESS] = rta_tb[IFA_LOCAL];


	switch (ifa->ifa_family) {
	case AF_INET:
		memcpy(&NET_GET_ADDR_IPv4(&vifa->local), RTA_DATA(rta_tb[IFA_LOCAL]), alen);
		if (rta_tb[IFA_ADDRESS])
			memcpy(&NET_GET_ADDR_IPv4(&vifa->prefix), RTA_DATA(rta_tb[IFA_ADDRESS]), alen);
		else
			vifa->prefix = vifa->local;
		break;
	case AF_INET6:
		memcpy(&NET_GET_ADDR_IPv6(&vifa->local), RTA_DATA(rta_tb[IFA_LOCAL]), alen);
		if (rta_tb[IFA_ADDRESS])
			memcpy(&NET_GET_ADDR_IPv6(&vifa->prefix), RTA_DATA(rta_tb[IFA_ADDRESS]), alen);
		else
			vifa->prefix = vifa->local;
		break;
	}
*/
	return 0;
}

static int store_linkinfo(struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	int i;
	struct ifinfomsg *ifi = NLMSG_DATA(n);
	struct rtattr *rta[IFLA_MAX+1];

	if (n->nlmsg_type != RTM_NEWLINK)
		return 0;
	if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifi))) {
		errno = 0;
		return -1;
	}

	parse_rtattr(rta, IFLA_MAX, IFLA_RTA(ifi), IFLA_PAYLOAD(n));
	if (rta[IFLA_IFNAME] == NULL || rta[IFLA_MTU] == NULL)
		return 0;

	i = ifi->ifi_index;
	if (i >= MAX_INTERFACES)
		return 0;
	ifap[i].flags = ifi->ifi_flags;
	ifap[i].mtu = *(unsigned*)RTA_DATA(rta[IFLA_MTU]);
	return 0;
}

void
linux_init()
{
	static int once = FALSE;
	struct rtnl_handle rth;

	if (once)
		return;

	if (FAILED(rtnl_open(&rth, 0)))
		return;

	/* Read list of links */

	if (FAILED(rtnl_wilddump_request(&rth, AF_UNSPEC, RTM_GETLINK)))
		return;

	if (FAILED(rtnl_dump_filter(&rth, store_linkinfo, NULL)))
		return;

	/* Now get protocol information */

	if (FAILED(rtnl_wilddump_request(&rth, AF_INET, RTM_GETADDR)))
		return;

	if (FAILED(rtnl_dump_filter(&rth, store_addrinfo, NULL)))
		return;

#ifdef	USE_IPV6
	if (FAILED(rtnl_wilddump_request(&rth, AF_INET6, RTM_GETADDR)))
		return;

	if (FAILED(rtnl_dump_filter(&rth, store_addrinfo, NULL)))
		return;
#endif	/* USE_IPV6 */
	rtnl_close(&rth);
	once = TRUE;
}

#endif	/* linux */

#ifdef	sun

struct if_nameindex *
if_nameindex()
{
	int fd,n = 0;
	struct ifconf ifc;
	struct IFREQ *p,*last;
	static char buffer[MAX_INTERFACES * sizeof(struct IFREQ)
		+ sizeof(struct ifconf)];

	fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(NULL);
	ifc.ifc_buf = buffer;
	ifc.ifc_len = sizeof(buffer);
	if (FAILED(ioctl(fd,_SIOCGIFCONF,(caddr_t) &ifc))) {
		close(fd);
		return(NULL);
	}
	close(fd);
	last = ((struct IFREQ *) ifc.ifc_req) + ifc.ifc_len;
	for (p = (struct IFREQ *) ifc.ifc_req;p < last; p++) {
		if (strncmp(p->IF_NAME,"",IFNAMSIZ) == 0)
			continue;
		ifs[n].if_name = p->IF_NAME;
		ifs[n].if_index = n + 1;
		n++;
	}
	ifs[n].if_index = 0;
	ifs[n].if_name = NULL;
	return(ifs);
}

#endif	/* sun */
#endif	/* __FreeBSD__ */

#if	(defined(__FreeBSD__) || defined(sun))

void
if_freenameindex(struct if_nameindex *p)
{
}

#endif	/* (defined(__FreeBSD__) || defined(sun)) */

unsigned int
if_nametoindex(const char *ifname)
{
	struct if_nameindex *p,*q;
	unsigned int index = 0;

	p = if_nameindex();
	if (p == NULL)
		return(index);
	for (q = p;q->if_name != NULL; q++)
		if (strncmp(ifname,q->if_name,IFNAMSIZ) == 0) {
			index = q->if_index;
			break;
		}
	if_freenameindex(p);
	return(index);
}

char *
if_indextoname(unsigned int ifindex,char *ifname)
{

	struct if_nameindex *p,*q;
	char *name = NULL;

	p = if_nameindex();
	if (p == NULL)
		return(name);
	for (q = p;q->if_name != NULL; q++)
		if (ifindex == q->if_index) {
			strncpy(ifname,q->if_name,IFNAMSIZ);
			name = ifname;
			break;
		}
	if_freenameindex(p);
	return(name);
}

#ifdef	sun

const char *
inet_ntop(int af,const void *src,char *dst,size_t size)
{
	switch (af) {
		case AF_INET:
			if (size < INET_ADDRSTRLEN)
				return(NULL);
			strncpy(dst,inet_ntoa(*(struct in_addr *) src),
				INET_ADDRSTRLEN);
			return(dst);
#ifdef	USE_IPV6
		case AF_INET6:
			if (size < INET6_ADDRSTRLEN)
				return(NULL);
			strncpy(dst,addr2ascii(AF_INET6,
				(struct in6_addr *) src,
				sizeof(struct in6_addr),NULL),
				INET6_ADDRSTRLEN);
			return(dst);
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

int
inet_pton(int af,const char *src,void *dst)
{
	switch (af) {
		case AF_INET:
			((struct in_addr *) dst)->s_addr = inet_addr(src);
			return((((struct in_addr *) dst)->s_addr == SYS_ERROR)
				? FALSE : 1);
#ifdef	USE_IPV6
		case AF_INET6:
			return((ascii2addr(AF_INET6,src,dst) == SYS_ERROR)
				? FALSE : 1);
#endif	/* USE_IPV6 */
		default:
			return(SYS_ERROR);
	}
}

struct hostent *
gethostbyname2(const char *name,int af)
{
	switch (af) {
		case AF_INET:
			return(gethostbyname(name));
#ifdef	USE_IPV6
		case AF_INET6:
			return(hostname2addr(name,AF_INET6));
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
	return(NULL);
}

#endif	/* sun */

struct if_attributes *
if_attributes(unsigned int n)
{
#ifdef	__FreeBSD__

	unsigned int i,flags,mtu;
	struct if_attributes *l;
	struct sockaddr *s;
	char name[IFNAMSIZ];

	if (if_indextoname(n,name) == NULL)
		return(NULL);
	flags = if_indextoflags(n);
	mtu = if_indextomtu(n);
	l = ifap[--n];
	i = 0;
	for(s = l->addr; s != NULL; s = (++l)->addr) {
		l->flags = flags;
		l->mtu = mtu;
		switch (s->sa_family) {
			case AF_INET:
				l->prefix = PREFIX(((struct sockaddr_in *)
					prefixes[n][i])->sin_addr.s_addr);
				break;
#ifdef	USE_IPV6
			case AF_INET6:
				l->prefix = PREFIX(((struct sockaddr_in6 *)
					prefixes[n][i])->sin6_addr);
				break;
#endif	/* USE_IPV6 */
			default:
				l->prefix = 0;
				break;
		}
		i++;
	}
	return(ifap[n]);

#else	/* __FreeBSD__ */

	static struct if_attributes ret[2];

	ret[0].addr = if_indextoaddr(n,AF_UNSPEC);
	ret[0].prefix = if_indextoprefix(n);
	ret[0].flags = if_indextoflags(n);
	ret[0].mtu = if_indextomtu(n);
	ret[1].addr = NULL;

#if	(defined(sun) && defined(USE_IPV6))
	if (ret[0].addr->sa_family == AF_INET6) {
		int fd;
		struct v6preinforeq ifr6;

		if (if_indextoname(n,ifr6.v6pr_name) == NULL)
			return(ret);
		fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
		if (FAILED(fd))
			return(ret);
		if (FAILED(ioctl(fd,SIOCGIFV6PREINFO,(caddr_t) &ifr6))) {
			close(fd);
			return(ret);
		}
		close(fd);
		if (!(ifr6.v6pr_flags & LV6IF_ADDRESSED))
			ret[0].addr = NULL;
	}
#endif	/* (defined(sun) && defined(USE_IPV6)) */

	return(ret);

#endif 	/* __FreeBSD__ */
}

unsigned int
if_addrtoindex(struct sockaddr *s)
{
	struct if_nameindex *p,*q;
	struct if_attributes *l;
	struct sockaddr *s2;

	p = if_nameindex();
	if (p == NULL)
		return(0);
	for (q = p;q->if_name != NULL; q++) {
		l = if_attributes(q->if_index);
		for(s2 = l->addr; s2 != NULL; s2 = (++l)->addr)
			if (net_sockaddr_equal(s,s2)) {
				if_freenameindex(p);
				return(q->if_index);
			}
	}
	if_freenameindex(p);
	return(0);
}

struct sockaddr *
if_indextoaddr(unsigned int index,int af)
{
#ifdef	__FreeBSD__

	struct sockaddr *s;
	struct if_attributes *l;

	l = if_attributes(index);
	/* FIX: choose BEST address ? */
	for(s = l->addr; s != NULL; s = (++l)->addr)
		if ((af == AF_UNSPEC) || (af == s->sa_family))
			return(s);
	return(NULL);

#else	/* __FreeBSD__ */

#ifdef	linux

	linux_init();
	if (index >= MAX_INTERFACES)
		return(NULL);
	return(ifap[index].addr);

#else	/* linux */

	int fd;
	static struct ifreq ifr;
#ifdef	USE_IPV6
	static struct IFREQADDR ifr6;
	static struct sockaddr_in6 sin;

	if (if_indextoname(index,ifr6.IF_ADDRNAME) == NULL)
		return(NULL);
	fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(NULL);
	if (FAILED(ioctl(fd,_SIOCGIFADDR,(caddr_t) &ifr6))) {
		close(fd);
		return(NULL);
	}
	close(fd);
	if (!IN6_ARE_ADDR_EQUAL(&ifr6.IF_ADDRADDR,&in6addr_any)) {
		if ((af != AF_UNSPEC) && (af != AF_INET6))
			return(NULL);
		NET_SOCKADDR_IPv6(&sin,ifr6.IF_ADDRADDR);
		return((struct sockaddr *) &sin);
	}
#endif	/* USE_IPV6 */
	if (if_indextoname(index,ifr.ifr_name) == NULL)
		return(NULL);
	fd = socket(AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(NULL);
	if (FAILED(ioctl(fd,SIOCGIFADDR,(caddr_t) &ifr))) {
		close(fd);
		return(NULL);
	}
	close(fd);
	if ((af != AF_UNSPEC) && (af != AF_INET))
		return(NULL);
	return((struct sockaddr *) &ifr.ifr_addr);

#endif	/* linux */

#endif 	/* __FreeBSD__ */
}

#ifndef	__FreeBSD__

static
int
if_indextoprefix(unsigned int n)
{
#ifdef	linux

	if (n >= MAX_INTERFACES)
		return(0);
	if (ifap[n].addr == NULL)
		return(0);
	return(ifap[n].prefix);

#else	/* linux */

	int fd,ret = 0;
	struct sockaddr *s;
	struct ifreq ifr;
#ifdef	USE_IPV6
	struct IFREQMASK ifr6;
#endif	/* USE_IPV6 */

	s = if_indextoaddr(n,AF_UNSPEC);
	if (s == NULL)
		return(ret);
	fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(ret);
	switch (s->sa_family) {
		case AF_INET:
			if_indextoname(n,ifr.ifr_name);
			if (FAILED(ioctl(fd,SIOCGIFNETMASK,(caddr_t) &ifr)))
				break;
			ret = PREFIX(((struct sockaddr_in *) &ifr.ifr_addr)
                                ->sin_addr.s_addr);
			break;
#ifdef	USE_IPV6
		case AF_INET6:
			if_indextoname(n,ifr6.IF_MASKNAME);
			if (FAILED(ioctl(fd,SIOCGIFNETMASK6,(caddr_t) &ifr6)))
				break;
			ret = PREFIX6(ifr6.IF_MASKADDR);
			break;
#endif	/* USE_IPV6 */
		default:
			break;
	}
	close(fd);
	return(ret);

#endif	/* linux */
}

#endif	/* __FreeBSD__ */

static
unsigned int
if_indextoflags(unsigned int n)
{
#ifdef	linux

	if (n >= MAX_INTERFACES)
		return(0);
	if (ifap[n].addr == NULL)
		return(0);
	return(ifap[n].flags);

#else	/* linux */

	int fd,ret = 0;
	struct ifreq ifr;

	fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(ret);
	if_indextoname(n,ifr.ifr_name);
	if (!FAILED(ioctl(fd,SIOCGIFFLAGS,(caddr_t) &ifr)))
		ret = ifr.ifr_flags;
	close(fd);
	return(ret);

#endif	/* linux */
}

static
unsigned int
if_indextomtu(unsigned int n)
{
#ifdef	linux

	if (n >= MAX_INTERFACES)
		return(0);
	if (ifap[n].addr == NULL)
		return(0);
	return(ifap[n].mtu);

#else	/* linux */

	int fd,ret = 0;
	struct ifreq ifr;

	fd = socket(_AF_INET,SOCK_DGRAM,PF_UNSPEC);
	if (FAILED(fd))
		return(ret);
	if_indextoname(n,ifr.ifr_name);
	if (!FAILED(ioctl(fd,SIOCGIFMTU,(caddr_t) &ifr)))
		ret = ifr.ifr_metric;
	close(fd);
	return(ret);

#endif	/* linux */
}

#ifndef	linux

static
unsigned int
prefix(unsigned char *p,unsigned int n)
{
	unsigned int i,sum = 0;

	for (i = 0;i < n; i++)
		if (BSET_ISSET(i,p))
			sum++;
	return(sum);
}

#endif	/* linux */

const char *
net_inaddr_print(const char *s,int af)
{
	static char buffer[MAX_ADDRSTRLEN];

	switch (af) {
		case AF_INET:
			return(inet_ntop(AF_INET,
				(struct in_addr *) s,
				buffer,sizeof(buffer)));
#ifdef	USE_IPV6
		case AF_INET6:
			return(inet_ntop(AF_INET6,
				(struct in6_addr *) s,
				buffer,sizeof(buffer)));
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

const char *
net_inaddr_host(const char *s,int af)
{
	struct hostent *hp;

	switch (af) {
		case AF_INET:
			hp = gethostbyaddr(s,sizeof(struct in_addr),AF_INET);
			if (hp == NULL)
				return(net_inaddr_print(s,af));
			return(hp->h_name);
#ifdef	USE_IPV6
		case AF_INET6:
			hp = gethostbyaddr(s,sizeof(struct in6_addr),AF_INET6);
			if (hp == NULL)
				return(net_inaddr_print(s,af));
			return(hp->h_name);
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

int
net_sockaddr_equal(const struct sockaddr *x,const struct sockaddr *y)
{ 
	if (x->sa_family != y->sa_family)
		return(FALSE);
	switch (x->sa_family) {
		case AF_INET:
			if (!IN_ARE_ADDR_EQUAL(
					&((struct sockaddr_in *) x)->sin_addr,
					&((struct sockaddr_in *) y)->sin_addr))
				return(FALSE);
			return(((struct sockaddr_in *) x)->sin_port ==
				((struct sockaddr_in *) y)->sin_port);
#ifdef	USE_IPV6
		case AF_INET6:
			if (!IN6_ARE_ADDR_EQUAL(
					&((struct sockaddr_in6 *) x)->sin6_addr,
					&((struct sockaddr_in6 *) y)->sin6_addr))
				return(FALSE);
			return(((struct sockaddr_in6 *) x)->sin6_port ==
				((struct sockaddr_in6 *) y)->sin6_port);
#endif	/* USE_IPV6 */
		default:
			return(FALSE);
	}
}

const char *
net_sockaddr_print(const struct sockaddr *s)
{
	static char buffer[MAX_ADDRPORTSTRLEN],buffer2[MAX_ADDRSTRLEN];

	switch (s->sa_family) {
		case AF_INET:
			sprintf(buffer,"%s/%hu",inet_ntop(AF_INET,
				&((struct sockaddr_in *) s)->sin_addr,
				buffer2,sizeof(buffer2)),
				ntohs(((struct sockaddr_in *) s)->sin_port));
			return(buffer);
#ifdef	USE_IPV6
		case AF_INET6:
			sprintf(buffer,"%s/%hu",inet_ntop(AF_INET6,
				&((struct sockaddr_in6 *) s)->sin6_addr,
				buffer2,sizeof(buffer2)),
				ntohs(((struct sockaddr_in6 *) s)->sin6_port));
			return(buffer);
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

int
net_addr_equal(const net_addr *x,const net_addr *y)
{
	if (NET_UNEQUAL_TYPE(x,y))
		return(FALSE);
	switch (NET_GET_TYPE(x)) {
		case NET_ADDR_IPv4:
			return(IN_ARE_ADDR_EQUAL(&NET_GET_ADDR_IPv4(x),
				&NET_GET_ADDR_IPv4(y)));
		case NET_ADDR_UDP_IPv4:
			return(memcmp(&NET_GET_ADDR_UDP_IPv4(x),
				&NET_GET_ADDR_UDP_IPv4(y),
				sizeof(NET_GET_ADDR_UDP_IPv4(x))) == 0);
#ifdef	USE_IPV6
		case NET_ADDR_IPv6:
			return(IN6_ARE_ADDR_EQUAL(&NET_GET_ADDR_IPv6(x),
				&NET_GET_ADDR_IPv6(y)));
		case NET_ADDR_UDP_IPv6:
			return(memcmp(&NET_GET_ADDR_UDP_IPv6(x),
				&NET_GET_ADDR_UDP_IPv6(y),
				sizeof(NET_GET_ADDR_UDP_IPv6(x))) == 0);
#endif	/* USE_IPV6 */
		default:
			return(FALSE);
	}
}

int
net_addr_is_multicast(const net_addr *addr)
{
	switch (NET_GET_TYPE(addr)) {
		case NET_ADDR_IPv4:
			return(IN_IS_ADDR_MULTICAST(
				&NET_GET_ADDR_IPv4(addr)));
		case NET_ADDR_UDP_IPv4:
			return(IN_IS_ADDR_MULTICAST(
				&NET_GET_ADDR_UDP_IPv4(addr).sin_addr));
#ifdef	USE_IPV6
		case NET_ADDR_IPv6:
			return(IN6_IS_ADDR_MULTICAST(
				&NET_GET_ADDR_IPv6(addr)));
		case NET_ADDR_UDP_IPv6:
			return(IN6_IS_ADDR_MULTICAST(
				&NET_GET_ADDR_UDP_IPv6(addr).sin6_addr));
#endif	/* USE_IPV6 */
		default:
			return(FALSE);
	}
}

net_addr *
net_addr_ip(net_addr *addr)
{
	static net_addr addr2;

	switch (NET_GET_TYPE(addr)) {
		case NET_ADDR_UDP_IPv4:
			NET_SET_ADDR_IPv4(&addr2,
				NET_GET_ADDR_UDP_IPv4(addr).sin_addr);
			return(&addr2);
#ifdef	USE_IPV6
		case NET_ADDR_UDP_IPv6:
			NET_SET_ADDR_IPv6(&addr2,
				NET_GET_ADDR_UDP_IPv6(addr).sin6_addr);
			return(&addr2);
#endif	/* USE_IPV6 */
		default:
			addr2 = *addr;
			return(&addr2);
	}
}

int
net_addr_assign(net_addr *addr,const struct sockaddr *s)
{
	switch (s->sa_family) {
		case AF_INET:
			NET_SET_ADDR_IPv4(addr,
				((struct sockaddr_in *) s)->sin_addr);
			return(TRUE);
#ifdef	USE_IPV6
		case AF_INET6:
			NET_SET_ADDR_IPv6(addr,
				((struct sockaddr_in6 *) s)->sin6_addr);
			return(TRUE);
#endif	/* USE_IPV6 */
		default:
			return(FALSE);
	}
}

int
net_addr_assign_udp(net_addr *addr,const struct sockaddr *s)
{
	switch (s->sa_family) {
		case AF_INET:
			NET_SET_ADDR_UDP_IPv4(addr,*(struct sockaddr_in *) s);
			return(TRUE);
#ifdef	USE_IPV6
		case AF_INET6:
			NET_SET_ADDR_UDP_IPv6(addr,*(struct sockaddr_in6 *) s);
			return(TRUE);
#endif	/* USE_IPV6 */
		default:
			return(FALSE);
	}
}

/*	Given character string containing host name, do DNS lookup(s)
 *	to convert to net_addr.
 */
int
net_addr_ascii(net_addr *addr,const char *str)
{
	struct hostent *hp;
	struct in_addr in;
#ifdef	USE_IPV6
	struct in6_addr in6;
#endif	/* USE_IPV6 */

	if (inet_pton(AF_INET,str,&in) == 1) {
		NET_SET_ADDR_IPv4(addr,in);
		return(TRUE);
	}
	hp = gethostbyname2(str,AF_INET);
	if (hp != NULL) {
		memcpy(&in,hp->h_addr,hp->h_length);
		NET_SET_ADDR_IPv4(addr,in);
		return(TRUE);
	}
#ifdef	USE_IPV6
	if (inet_pton(AF_INET6,str,&in6) == 1) {
		NET_SET_ADDR_IPv6(addr,in6);
		return(TRUE);
	}
	hp = gethostbyname2(str,AF_INET6);
	if (hp != NULL) {
		memcpy(&in6,hp->h_addr,hp->h_length);
		NET_SET_ADDR_IPv6(addr,in6);
		return(TRUE);
	}
#endif	/* USE_IPV6 */
	return(FALSE);
}

/*	Make printable numeric ("dotted decimal") format from net-addr.
 */
const char *
net_addr_print(const net_addr *addr)
{
	static char buffer[MAX_ADDRPORTSTRLEN],buffer2[MAX_ADDRSTRLEN];

	switch(NET_GET_TYPE(addr)) {
		case NET_ADDR_IPv4:
			return(inet_ntop(AF_INET,&NET_GET_ADDR_IPv4(addr),
				buffer,sizeof(buffer)));
		case NET_ADDR_UDP_IPv4:
			sprintf(buffer,"%s/%hu",inet_ntop(AF_INET,
				&NET_GET_ADDR_UDP_IPv4(addr).sin_addr,
				buffer2,sizeof(buffer2)),
				ntohs(NET_GET_ADDR_UDP_IPv4(addr).sin_port));
			return(buffer);
#ifdef	USE_IPV6
		case NET_ADDR_IPv6:
			return(inet_ntop(AF_INET6,&NET_GET_ADDR_IPv6(addr),
				buffer,sizeof(buffer)));
		case NET_ADDR_UDP_IPv6:
			sprintf(buffer,"%s/%hu",inet_ntop(AF_INET6,
				&NET_GET_ADDR_UDP_IPv6(addr).sin6_addr,
				buffer2,sizeof(buffer2)),
				ntohs(NET_GET_ADDR_UDP_IPv6(addr).sin6_port));
			return(buffer);
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

/*	Map net_addr into host[/port] character string using inverse
 *	DNS lookup.
 */
const char *
net_addr_host(const net_addr *addr)
{
	static char buffer[MAX_HOSTPORTSTRLEN];
	struct hostent *hp;

	switch(NET_GET_TYPE(addr)) {
		case NET_ADDR_IPv4:
			hp = gethostbyaddr((char *) &NET_GET_ADDR_IPv4(addr),
				sizeof(NET_GET_ADDR_IPv4(addr)),AF_INET);
			if (hp == NULL)
				return(net_addr_print(addr));
			return(hp->h_name);
		case NET_ADDR_UDP_IPv4:
			hp = gethostbyaddr(
				(char *) &NET_GET_ADDR_UDP_IPv4(addr).sin_addr,
				sizeof(NET_GET_ADDR_UDP_IPv4(addr).sin_addr),
				AF_INET);
			if (hp == NULL)
				return(net_addr_print(addr));
			sprintf(buffer,"%s/%hu",hp->h_name,
				ntohs(NET_GET_ADDR_UDP_IPv4(addr).sin_port));
			return(buffer);
#ifdef	USE_IPV6
		case NET_ADDR_IPv6:
			hp = gethostbyaddr((char *) &NET_GET_ADDR_IPv6(addr),
				sizeof(NET_GET_ADDR_IPv6(addr)),AF_INET6);
			if (hp == NULL)
				return(net_addr_print(addr));
			return(hp->h_name);
		case NET_ADDR_UDP_IPv6:
			hp = gethostbyaddr(
				(char *) &NET_GET_ADDR_UDP_IPv6(addr).sin6_addr,
				sizeof(NET_GET_ADDR_UDP_IPv6(addr).sin6_addr),
				AF_INET6);
			if (hp == NULL)
				return(net_addr_print(addr));
			sprintf(buffer,"%s/%hu",hp->h_name,
				ntohs(NET_GET_ADDR_UDP_IPv6(addr).sin6_port));
			return(buffer);
#endif	/* USE_IPV6 */
		default:
			return(NULL);
	}
}

int
net_if_equal(const net_if *x,const net_if *y)
{
	if (NET_UNEQUAL_TYPE(x,y))
		return(FALSE);
	switch (NET_GET_TYPE(x)) {
		case NET_IF_UNKNOWN:
			return(TRUE);
		case NET_IF_ID:
			return(NET_GET_IF_ID(x) == NET_GET_IF_ID(y));
		case NET_IF_STRING:
			return(strcmp(NET_GET_IF_STRING(x),
				NET_GET_IF_STRING(y)) == 0);
		case NET_IF_PHY:
			return((NET_GET_IF_PHY_ID(x) == NET_GET_IF_PHY_ID(y))
				&& net_addr_equal(&NET_GET_IF_PHY_ADDR(x),
				&NET_GET_IF_PHY_ADDR(y)));
		case NET_IF_VIF:
			return((NET_GET_IF_VIF_ID(x) == NET_GET_IF_VIF_ID(y))
				&& net_addr_equal(&NET_GET_IF_VIF_ADDR(x),
				&NET_GET_IF_VIF_ADDR(y)));
		default:
			return(FALSE);
	}
}

const char *
net_if_print(const net_if *inf)
{
	static char buffer[MAX_ADDRSTRLEN];

	switch(NET_GET_TYPE(inf)) {
		case NET_IF_UNKNOWN:
			return("unknown");
		case NET_IF_ID:
			sprintf(buffer,"%d",NET_GET_IF_ID(inf));
			return(buffer);
		case NET_IF_STRING:
			return(NET_GET_IF_STRING(inf));
		case NET_IF_PHY:
			return(net_addr_print(&NET_GET_IF_PHY_ADDR(inf)));
		case NET_IF_VIF:
			sprintf(buffer,"%s vif %d",
				net_addr_print(&NET_GET_IF_VIF_ADDR(inf)),
				NET_GET_IF_VIF_ID(inf));
			return(buffer);
		default:
			return(NULL);
	}
}


/*	Determine default interface by doing inverse DNS lookup on
 *	my host name.
 */
net_addr *
net_get_default_interface()
{
	char name[MAXHOSTNAMELEN];
	static net_addr addr;

	if (FAILED(gethostname(name,MAXHOSTNAMELEN)))
		return(NULL);
	if (!net_addr_ascii(&addr,name))
		return(NULL);
	return(&addr);
}

#ifdef	linux

/*
 * libnetlink.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 */

#define MSG_TRUNC	0x20
#define MSG_DONTWAIT	0x40	/* Nonblocking io		 */

int rtnl_open(struct rtnl_handle *rth, unsigned subscriptions)
{
	int addr_len;

	memset(rth, 0, sizeof(rth));

	rth->fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (rth->fd < 0) {
		fprintf(stderr,"cannot open netlink socket: %m");
		return -1;
	}

	memset(&rth->local, 0, sizeof(rth->local));
	rth->local.nl_family = AF_NETLINK;
	rth->local.nl_groups = subscriptions;

	if (bind(rth->fd, (struct sockaddr*)&rth->local, sizeof(rth->local)) < 0) {
		fprintf(stderr,"cannot bind netlink socket: %m");
		goto failure;
	}
	addr_len = sizeof(rth->local);
	if (getsockname(rth->fd, (struct sockaddr*)&rth->local, &addr_len) < 0) {
		fprintf(stderr,"cannot getsockname: %m");
		goto failure;
	}
	if (addr_len != sizeof(rth->local)) {
		fprintf(stderr,"wrong address length %d", addr_len);
		goto failure;
	}
	if (rth->local.nl_family != AF_NETLINK) {
		fprintf(stderr,"wrong address family %d", rth->local.nl_family);
		goto failure;
	}
	rth->seq = 1;
	return 0;

failure:
	rth->fd = -1;
	return -1;
}

void rtnl_close(struct rtnl_handle *rth)
{
	close(rth->fd);
	rth->fd = -1;
}

void rtnl_set_junkfilter(struct rtnl_handle *rth,
			 int (*junk)(struct sockaddr_nl *,struct nlmsghdr *n),
			 int (**o_junk)(struct sockaddr_nl *,struct nlmsghdr *n))
{
	if (o_junk)
		*o_junk = rth->junk;
	rth->junk = junk;
}

int rtnl_dump_request(struct rtnl_handle *rth, int type, void *req, int len)
{
	int err;
	struct nlmsghdr nlh;
	struct sockaddr_nl nladdr;
	struct iovec iov[2] = { { &nlh, sizeof(nlh) }, { req, len } };
	struct msghdr msg = {
		(void*)&nladdr, sizeof(nladdr),
		iov,	2,
		NULL,	0,
		0
	};

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;

	nlh.nlmsg_len = NLMSG_LENGTH(len);
	nlh.nlmsg_type = type;
	nlh.nlmsg_flags = NLM_F_ROOT|NLM_F_REQUEST;
	nlh.nlmsg_pid = 0;
	nlh.nlmsg_seq = rth->dump = ++rth->seq;

	err = sendmsg(rth->fd, &msg, 0);
	if (err == nlh.nlmsg_len)
		return 0;
	return -1;
}


int rtnl_wilddump_request(struct rtnl_handle *rth, int family, int type)
{
	struct rtgenmsg g;

	memset(&g, 0, sizeof(g));
	g.rtgen_family = family;

	return rtnl_dump_request(rth, type, (void*)&g, sizeof(g));
}

/* As fast as possible flush queue.
   A part of information is lost in any case, so that
   the best that we can make now is to ignore burst
   and try to recover.
 */
void rtnl_flush(struct rtnl_handle *rth)
{
	char buf[8];

	while (recv(rth->fd, buf, sizeof(buf), MSG_DONTWAIT) >= 0 ||
	       errno != EAGAIN)
		/* NOTHING */;

	sleep(1);

	while (recv(rth->fd, buf, sizeof(buf), MSG_DONTWAIT) >= 0 ||
	       errno != EAGAIN)
		/* NOTHING */;
}

int rtnl_dump_filter(struct rtnl_handle *rth,
		     int (*filter)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
		     void *arg1)
{
	char	buf[8192];
	struct sockaddr_nl nladdr;
	struct iovec iov = { buf, sizeof(buf) };

	while (1) {
		int status;
		struct nlmsghdr *h;

		struct msghdr msg = {
			(void*)&nladdr, sizeof(nladdr),
			&iov,	1,
			NULL,	0,
			0
		};

		status = recvmsg(rth->fd, &msg, 0);

		if (status < 0) {
			int s_errno = errno;
			if (errno == EINTR)
				continue;

			rtnl_flush(rth);
			errno = s_errno;
			return -1;
		}
		if (status == 0) {
			fprintf(stderr,"EOF on netlink");
			errno = EINVAL;
			return -1;
		}
		if (msg.msg_namelen != sizeof(nladdr)) {
			fprintf(stderr,"sender address length == %d", msg.msg_namelen);
			errno = EINVAL;
			return -1;
		}
		if (nladdr.nl_pid) {
			fprintf(stderr,"message from %d", nladdr.nl_pid);
			continue;
		}

		h = (struct nlmsghdr*)buf;
		while (NLMSG_OK(h, status)) {
			int err;

			if (h->nlmsg_pid != rth->local.nl_pid ||
			    h->nlmsg_seq != rth->dump) {
				if (rth->junk)
					rth->junk(&nladdr, h);
				goto skip_it;
			}

			if (h->nlmsg_type == NLMSG_DONE)
				return 0;
			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(h);
				if (h->nlmsg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr))) {
					fprintf(stderr,"ERROR truncated");
					errno = EMSGSIZE;
				} else {
					if (err->error == 0)
						return 0;
					fprintf(stderr,"not OK\n");
					errno = -err->error;
				}
				return -1;
			}
			err = filter(&nladdr, h, arg1);
			if (err < 0)
				return err;

skip_it:
			h = NLMSG_NEXT(h, status);
		}

		if (msg.msg_flags & MSG_TRUNC) {
			errno = EMSGSIZE;
			return -1;
		}
		if (status)
			fprintf(stderr,"!!!Remnant of size %d", status);
	}
}


int rtnl_ask(struct rtnl_handle *rtnl, struct iovec *iov, int iovlen,
	     struct rtnl_dialog *d, char *buf, int len)
{
	int i;
	struct nlmsghdr *n = iov[0].iov_base;
	struct sockaddr_nl nladdr;
	struct msghdr msg = {
		(void*)&nladdr, sizeof(nladdr),
		iov,	iovlen,
		NULL,   0,
		0
	};

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;

	n->nlmsg_seq = d->seq = ++rtnl->seq;
	n->nlmsg_flags |= NLM_F_ACK|NLM_F_REQUEST;
	n->nlmsg_len = 0;

	for (i=0; i<iovlen-1; i++) {
		iov[i].iov_len = NLMSG_ALIGN(iov[i].iov_len);
		n->nlmsg_len += iov[i].iov_len;
	}
	n->nlmsg_len += iov[iovlen-1].iov_len;

	d->buf = d->ptr = buf;
	d->buflen = len;
	d->curlen = 0;

	i = sendmsg(rtnl->fd, &msg, 0);
	if (i == n->nlmsg_len)
		return 0;
	return -1;
}

struct nlmsghdr * rtnl_wait(struct rtnl_handle *rth, struct rtnl_dialog *d,
			    int *err)
{
	int status;
	struct nlmsghdr *n;

	while (1) {
		if (d->curlen == 0) {
			struct iovec iov = { (void*)d->buf, d->buflen };
			struct msghdr msg = {
				(void*)&d->peer, sizeof(d->peer),
				&iov,	1,
				NULL,	0,
				0
			};

			status = recvmsg(rth->fd, &msg, 0);

			if (status < 0) {
				int s_errno = errno;
				if (errno == EINTR)
					continue;
				rtnl_flush(rth);
				errno = s_errno;
				*err = -1;
				return NULL;
			}
			if (status == 0) {
				fprintf(stderr,"EOF on netlink");
				*err = -1;
				errno = EINVAL;
				return NULL;
			}
			if (msg.msg_namelen != sizeof(d->peer)) {
				fprintf(stderr,"sender address length == %d\n", msg.msg_namelen);
				*err = -1;
				errno = EINVAL;
				return NULL;
			}
			if (d->peer.nl_pid) {
				fprintf(stderr,"message from %d", d->peer.nl_pid);
				continue;
			}
			d->curlen = status;
			d->ptr = d->buf;
		}

		n = (struct nlmsghdr*)d->ptr;
		status = d->curlen;

		while (NLMSG_OK(n, status)) {
			d->curlen -= NLMSG_ALIGN(n->nlmsg_len);
			d->ptr += NLMSG_ALIGN(n->nlmsg_len);
			
			if (n->nlmsg_pid != rth->local.nl_pid ||
			    n->nlmsg_seq != d->seq) {
				if (rth->junk)
					rth->junk(&d->peer, n);
				goto skip_it;
			}

			if (n->nlmsg_type == NLMSG_DONE) {
				*err = 0;
				return NULL;
			}
			if (n->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *e = (struct nlmsgerr*)NLMSG_DATA(n);
				if (n->nlmsg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr))) {
					fprintf(stderr,"ERROR truncated");
					errno = EINVAL;
				} else if (e->error == 0) {
					*err = 0;
					return NULL;
				} else {
					errno = -e->error;
					fprintf(stderr,"RTNETLINK error: %m");
					errno = -e->error;
				}
				*err = -1;
				return NULL;
			}
			return n;

skip_it:
			n = NLMSG_NEXT(n, status);
		}
		d->curlen = 0;
		d->ptr = d->buf;
	}
}

int rtnl_tell_iov(struct rtnl_handle *rth, struct iovec *iov, int iovlen)
{
	int err;
	char buf[4096];
	struct rtnl_dialog d;

	err = rtnl_ask(rth, iov, iovlen, &d, buf, sizeof(buf));
	if (err)
		return err;

	while (rtnl_wait(rth, &d, &err) != NULL)
		/* NOTHING */;

	return err;
}

int rtnl_tell(struct rtnl_handle *rth, struct nlmsghdr *n)
{
	struct iovec iov = { (void*)n, n->nlmsg_len };

	return rtnl_tell_iov(rth, &iov, 1);
}

int addattr32(struct nlmsghdr *n, int maxlen, int type, __u32 data)
{
	int len = RTA_LENGTH(4);
	struct rtattr *rta;
	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
		return -1;
	rta = (struct rtattr*)(((char*)n) + NLMSG_ALIGN(n->nlmsg_len));
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), &data, 4);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
	return 0;
}

int addattr_l(struct nlmsghdr *n, int maxlen, int type, void *data, int alen)
{
	int len = RTA_LENGTH(alen);
	struct rtattr *rta;
	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen)
		return -1;
	rta = (struct rtattr*)(((char*)n) + NLMSG_ALIGN(n->nlmsg_len));
	rta->rta_type = type;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), data, alen);
	n->nlmsg_len = NLMSG_ALIGN(n->nlmsg_len) + len;
	return 0;
}

int rta_addattr_l(struct rtattr *rta, int maxlen, int type, void *data, int alen)
{
	struct rtattr *subrta;
	int len = RTA_LENGTH(alen);

	if (RTA_ALIGN(rta->rta_len) + len > maxlen)
		return -1;
	subrta = (struct rtattr*)(((char*)rta) + RTA_ALIGN(rta->rta_len));
	subrta->rta_type = type;
	subrta->rta_len = len;
	memcpy(RTA_DATA(subrta), data, alen);
	rta->rta_len = NLMSG_ALIGN(rta->rta_len) + len;
	return 0;
}

int parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta, int len)
{
	memset(tb, 0, sizeof(struct rtattr*)*(max+1));

	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta,len);
	}
	if (len) {
		fprintf(stderr,"!!!Deficit %d: type=%d len=%d\n", len,
		       rta->rta_type, rta->rta_len);
		return -EINVAL;
	}
	return 0;
}

#endif	/* linux */
